#!/usr/bin/python
import sys
import subprocess
from datetime import date
from datetime import timedelta
from docopt import docopt

######################################################
# the functions which invoke subprocess

def get_find_results( directory, option = "" ):
    cmd =""" find %s %s 2>/dev/null"""%( directory, option )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )  
    (results,err) = p.communicate()
    results = results.split()
    return results


def ls_a_folder( directory, option = ""  ):
    # note: if directory does not exist, this function returns an empty list.
    cmd =""" ls %s %s 2>/dev/null"""%( directory, option )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )  
    (results,err) = p.communicate()
    results = results.split()

    return results


def get_run_dry_result( source, dirBackup):
    # call rsync --dry-run

    cmd =""" rsync --verbose --compress \
                   --archive --recursive \
                   --itemize-changes --delete \
                    --dry-run  %s %s 
    """%( source, dirBackup )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )  
    (results,err) = p.communicate()
    results = results.split("\n") 
    results = results[1:-3]      
    return results


def do_rsync( src, mydir  ):
    cmd =""" rsync --compress --archive  --recursive --delete   \
        %s %s """  %( src,  mydir  )
     
    subprocess.check_call( cmd, shell=True,  stdout=subprocess.PIPE )
    # to use check_call or Popen? 

    
def do_cat( fName ):
    cmd =""" cat  %s """  %( fName )
    subprocess.check_call( cmd, shell=True  )


def getAbsolutePath_from_relative( relative_path ):
    # convert relative path to absolute path
    # it use 'dirname' and 'basename'
    cmd=""" echo "$(cd "$(dirname %s)"; pwd)/$(basename %s)"  """ \
         %( relative_path, relative_path)
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate() 
    abs_path=result.strip()
    return abs_path 


def getDateTime_info( ):
    cmd = 'date "+%Y%m%d%H%M"'
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (dateTimeInfo,err) = p.communicate() 
    dateTimeInfo = dateTimeInfo.strip()
    return dateTimeInfo


def getCurrentPath( ):
    cmd="pwd"
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate() 
    currentPATH=result.strip()
    return currentPATH 


def makePatch( old, new , patch_fileName ):
    # note -u option give us time stamp
    cmd = "diff -u %s %s > %s" %( old, new, patch_fileName )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate()
    

def isDifferent( old, new ):
    cmd = "diff -u %s %s" %( old, new )
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate() 
     
    if result != "":
        return True
    else:
        return False




#### the following functions do not have return

def applyPatch_Reverse( fName, patch_fName ):
    
    # option "--reject-file -" or "-r -" is to discard .rej files.
    cmd = "patch -r - -R --ignore-whitespace %s < %s " %(fName, patch_fName )
    
    p = subprocess.Popen( cmd, shell=True,  stdout=subprocess.PIPE )
    (result,err) = p.communicate()

    # print 'debug result = ', result 
    # print 'debug err = ', err 

   
    

def make_a_fold( folderName ):
    cmd = "mkdir -p %s" %( folderName )
    subprocess.check_call( cmd, shell=True,  stdout=subprocess.PIPE )

def remove_a_folder( folderName ):
    cmd = "rm -r %s" %( folderName )
    subprocess.check_call( cmd, shell=True,  stdout=subprocess.PIPE )
    pass


def do_touch( dateTime, fName ):
    # change modification time of a file
    cmd = "touch  -m -t %s %s" %( dateTime, fName )
    subprocess.check_call( cmd, shell=True,  stdout=subprocess.PIPE )
    pass

##### handy functions (no subprocess )

def get_fileName( _fPath ):
    # extract file name from a path
    
    if( _fPath.find("/") == -1 ):
        return _fPath

    N = len(_fPath)
    idx = 0 
    for i in range( len(_fPath) ):
        if( _fPath[ -N + i ] == '/' ):
            idx = i 
              
    return _fPath[ idx+1: ] 


def get_dirName( _fileName ):
     
    N = len(_fileName)
    idx = 0 
    for i in range( len(_fileName) ):
        if( _fileName[ -N + i ] == '/' ):
            idx = -N + i 
              
    return _fileName[ :idx ] 

#=======================================================================




class RSYNC( ):
    """ to use rsync to do back up and manage version history """
    
    def __init__( self ):
        
        # the backupDir is the topmost dir for backup
        # sources can be either dir or files.
        # they can be assigned by set_main_backup_folder 
        # and add_source methods
        self.backupDir = ""
        self.sources = []

        self.upperDays = sys.maxint 
        # keep version upto XX days.
        # it can be changed by set_upper_days()

        # for processing
        self.modified_files = []
        self.deleted_files = []
        self.history_versions = []

        # fixed dirname
        self.tempDir=".temp"        
        self.historyDir=".history"  
        pass

    def process( self ):
        ''' the main backup process is done here'''
        self.get_deleted_files()
        if( len( self.deleted_files ) > 0 ):
            self.archieve_deleted_files_entrieCopy()

        self.get_modified_files()
        if( len( self.modified_files )>0 ):
            self.make_patches_for_modified_files()
        
        self.update_backupDir()  
        print "backup done"
        pass

    ###################################################
    # methods related to user setting

    def read_in_backup_plan( self, backup_plan ):
        ''' read in a backup instruction and then process it '''
        with open(backup_plan, "r" ) as f:
            lines = f.readlines()

        for line in lines:
            if( line[0] == "#" or len(line) == 0 ) : continue

            if line.find( "main_backup_folder" ) != -1:
                line = line.strip() 
                main_backup_folder = line.split(":")[-1]
                main_backup_folder = main_backup_folder.strip()
                self.set_main_backup_folder( main_backup_folder )

            if line.find( "add_source" ) != -1 :
                line = line.strip() 
                source = line.split(":")[-1]
                source = source.strip()
                self.add_source( source )
            
            if line.find( "keep_version" ) != -1 :
                line = line.strip() 
                duration = line.split(":")[-1]     
                duration = int( duration )  
                self.set_upper_days( duration )  
                self.manage_history_version()
        
        self.process()            
        pass


    def set_upper_days( self, days ):
        self.upperDays = days     
        

    def set_main_backup_folder( self, backupDir ):
        ''' assign self.backupDir'''
        if( backupDir[-1] == "/" ):
            backupDir = backupDir[ :-1]
        self.backupDir = backupDir 



    def add_source( self, fileName_path ):
        ''' add to self.sources'''
        self.sources.append( fileName_path )

    ###################################################
    # methods related to retrieve info
    
    def get_modified_files( self ):
        ''' this function return a list containing the filenames for 
            modified files.
        '''

        ## warning message
        if self.backupDir == "":
            print "no backup folder is set"
            print "use set_main_backup_folder() "
            return 

        ## warning message    
        if len( self.sources ) == 0 :
            print "no source is set"
            print "use add_source() "
            return 

        for source in self.sources:
            dry_run_results = get_run_dry_result( source, self.backupDir )
            self.get_modified_file_list( dry_run_results )
             
        return self.modified_files
          

    def get_deleted_files( self ):
        ''' this function return a list containing the filenames for 
            deleted files.
        '''
        ## warning message
        if self.backupDir == "":
            print "no backup folder is set"
            print "use set_main_backup_folder() "
            return 

        ## warning message    
        if len( self.sources ) == 0 :
            print "no source is set"
            print "use add_source() "
            return 

        for source in self.sources:
            dry_run_results = get_run_dry_result( source, self.backupDir ) 
            self.get_deleted_file_list( dry_run_results )
            
        return  self.deleted_files
       

    def get_modified_file_list( self, dry_run_results ):
        ''' rsync shows >f for modified files, and we can
            use this method to extract these modified files.
        '''
        STATUSLEN = 12
        for result in dry_run_results:
            if( result[:2] == ">f" and (result.find("+") == -1 ) ):
                filename = result[ STATUSLEN: ] 
                if( filename not in self.modified_files ):
                    self.modified_files.append( filename )

    def get_deleted_file_list( self, dry_run_results ):
        ''' rsync shows deleting for deleted files, and we can
            use this method to extract these deleted files.
        '''
        STATUSLEN = 12
        for result in dry_run_results:
            if(  result.find("deleting") != -1  ):
                filename = result[ STATUSLEN: ] 
                self.deleted_files.append( filename )



    ###################################################
    # methods related to do archieve and patch  


    def update_backupDir( self ):
        src = ""
        for source in self.sources:
            src += source + " "

        do_rsync( src, self.backupDir )
   

    def archieve_deleted_files_entrieCopy( self ):
        ''' to backup the deleted files '''
        deletedFolder = self.backupDir + "/" + self.historyDir + "/" + \
                        getDateTime_info() + "/deleted/"
        
        for source in self.deleted_files:
            dirName = get_dirName( deletedFolder + source )
            make_a_fold( dirName )
            src = self.backupDir+"/"+ source    
            do_rsync( src, dirName )    
        
        pass


    def archieve_modified_files_entrieCopy( self ):
        ''' currently not in used...''' 
        archiveFolder = self.backupDir + "/" + self.historyDir + "/" + \
                        getDateTime_info() + "/changed/"
        
        make_a_fold( archiveFolder )                
        src = ""
        for source in self.modified_files:
            src += self.backupDir+"/"+ source + " "
        
        do_rsync( src, archiveFolder )    
        pass


    def make_patches_for_modified_files( self ):
        # note: rsync uses modified date to compare
        # so even the content doesn't change, it is possible
        # to be considered changed file.


        patchFolder = self.backupDir + "/" + self.historyDir + "/" + \
                        getDateTime_info() + "/patch/"
          
        for source in self.modified_files:
            old_file = self.backupDir+ "/" + source
            new_file = source 
            path_file = patchFolder + source + ".patch"

            if( isDifferent( old_file, new_file ) ):
                dirName = get_dirName( patchFolder + source )
                make_a_fold( dirName )
                makePatch( old_file, new_file , path_file )
        pass


    ###################################################
    # methods related to modified files


    def get_file_versions( self, fName ):
        ''' given a file name, it returns modifying history ''' 
        if( fName[:2] == "./" ):
            fName = fName[2:]

        versions = []
         
        historyDir = self.backupDir + "/" +  self.historyDir
        results = get_find_results( historyDir ) 
         
        if( len(results) == 0  ):
            print "no history folder yet"
            return versions

        idx1 = len( historyDir) + 1
        idx2 = idx1 + 12 
        for result in results:
            if( result.find( fName) != -1 and result.find( "patch") != -1 ):
                dateTime = result[ idx1 : idx2 ]
                versions.append( dateTime )
                
        
        if( len(versions) == 0  ):
            print "no version yet"
            return []

        # sort
        versions.sort( reverse=True )

        return versions


    def get_versions_short( self, fName, dateTime, toPrint=True ):
        '''  extract the subset of patch files, which can 
            recover fName to version of dateTime
        '''

        versions = self.get_file_versions( fName )  
         
        # check the input
        inputCorrect=False
        for version in versions:
            if dateTime == version:
                inputCorrect = True

        if( inputCorrect == False ): 
            print "%s is not available" %( dateTime )
            return []
        else:
            if( toPrint ):
                for ii in range( len(versions) ):
                     
                    if( dateTime == versions[ii] ):
                        pass
                        print "%s <== selected" %( dateTime )
                    else:
                        print versions[ii]
         

        # get the shortened versions
        versions_short = []
        toBreakNextTime = False
        for version in versions:
            if( dateTime != version ):
                versions_short.append( version )
            elif( dateTime == version ):
                toBreakNextTime = True

            if( toBreakNextTime ): break

        return versions_short
        

    def back_to_previous_version( self, fName, dateTime ):
        # todo: remove debugFlag. we will create testing function
        # somewhere else.

        debugFlag = False

        versions_short = self.get_versions_short( fName, dateTime, False )
        
        if( len(versions_short) == 0 ): return

 
        if( debugFlag ): 
            for version in versions_short:  print version # for debug
            do_cat( fName )  # for debug 


        print "recover %s to %s: " %(fName, dateTime)
        # apply patches   
        patchName = "" 
        for version in versions_short:
            
            patchName = self.backupDir + "/" + self.historyDir + "/" \
                + version + "/patch/" +  fName + ".patch"

            if( debugFlag ):     
                print "-----------------------"   # for debug
                print "patch version = ", version # for debug 
             
                     
            applyPatch_Reverse( fName, patchName )
            
            if( debugFlag ): 
                print "after apply %s" %version  # for debug
                do_cat( fName )                  # for debug

        do_touch( dateTime, fName ) # change the modification time.  
        do_cat( fName )
        pass



    def get_dateTime_from_pathfile( self, pathFile ):
        ''' currently, it is not in used.
            return is a string ex. "201901010005"
        '''
        with open( pathFile, "r" ) as f:
            line = f.readline()
            line = line.strip()
        items = line.split()
        year_month_day = items[2]
        hh_mm = items[3]
        hh_mm = hh_mm[:5]
        year_month_day = year_month_day.split("-")
        hh_mm = hh_mm.split(":")
        outStr = ""
        for xx in year_month_day: outStr += xx
        for xx in hh_mm: outStr += xx    
        
        return outStr 

         

    def view_previous_version( self, fName, dateTime ):

        debugFlag = False
        
        versions_short = self.get_versions_short( fName, dateTime, True )

        
        # clean the old temp working folder (seems not necessary)    
        # dirToRemove = self.backupDir + "/" + self.tempDir 
        # remove_a_folder( dirToRemove )

        # make a temp working folder
        tempDir = self.backupDir + "/" + self.tempDir 
        make_a_fold( tempDir )

        # make a copy  
        do_rsync( fName, tempDir )
       
        # deal with file name
        fNameTemp = get_fileName( fName )
        fNameTemp = self.backupDir + "/" + self.tempDir + "/" + fNameTemp
        
        if( debugFlag ):
            print "-----------------------"   
            print "latest version"            
            do_cat( fNameTemp )               

        # apply patches        
        for version in versions_short:
            
            patchName = self.backupDir + "/" + self.historyDir + "/" \
                + version + "/patch/" +  fName + ".patch"

            if( debugFlag ):    
                print "-----------------------"  # for debug
                print "patch version = ", version
                print "debug fNameTemp = ", fNameTemp
                    
            applyPatch_Reverse( fNameTemp, patchName )
            
            if( debugFlag ):    
                print "after apply %s" %version 
                do_cat( fNameTemp ) #for debug

        # print on the screen.    
        do_cat( fNameTemp )    
        pass    
         
    ###################################################
    # methods related to deleted files.

    def show_all_deleted_files( self ):
        ''' print out all deleted files in backup/history/version/deleted/ '''
        historyDir = self.backupDir + "/" + self.historyDir + "/"
        results = get_find_results( historyDir, "-type f" )
         
        for fName in results:
            if( fName.find("deleted/") != -1 ):
                fName = fName.replace( historyDir ,"" )
                fName = fName.replace( "/deleted/" ," " )
                print fName
        pass        

    def preview_a_deleted_file( self, fName, dateTime):
        
        if( fName[:2] == './' ): fName.replace("./", "" )

        dirName =  self.backupDir + "/" + self.historyDir + "/" + \
                   dateTime + "/" + "deleted"
        results = get_find_results( dirName, "-type f" )           
         
        # validate input of fName, then do cat 
        isFound = False
        for result in results:
            if ( result.find(fName) != -1 ):
                print "preview ", result, ": "
                do_cat( result )
                isFound = True

        # error message
        if ( isFound == False ):
            print "Please check the input again, the requested file does not exist."
         
        pass           

    def recover_a_deleted_file( self, fName, dateTime):
        

        if( fName[:2] == './' ): fName.replace("./", "" )
        
        dirName =  self.backupDir + "/" + self.historyDir + "/" + \
                   dateTime + "/" + "deleted"
        
        results = get_find_results( dirName, "-type f" )           
        
        historyDir = self.backupDir + "/" + self.historyDir + "/"

        # validate input of fName, then do rsync 
        isFound = False
        for result in results:
            if ( result.find(fName) != -1 ):
                source_dir = get_dirName( fName )
                do_rsync( result, source_dir ) 
                result = result.replace( historyDir ,"" )
                result = result.replace( "/deleted/" ," " )
                print "recover: %s" %result
                isFound = True

        # error message
        if ( isFound == False ):
            print "Please check the input again, the requested file does not exist."
         
        pass        


    ###################################################
    # methods related to recover a dir to a dateTime.

    def recover_a_dir( self, dirName, dateTime ):
        pass    
    

    ###################################################
    # methods for history versions

    def get_history_versions( self ):
        ''' get all the dirnames (dateTime) under history dir''' 
        historyDir = self.backupDir + "/" + self.historyDir + "/"
        self.history_versions = ls_a_folder( historyDir ) 
        return self.history_versions

    def manage_history_version( self ):
        self.get_history_versions()
        
        today = date.today()

        for version in self.history_versions:
            year   = int( version[  :4] )
            month  = int( version[ 4:6] ) 
            day    = int( version[ 6:8] )
            # hour   = int( version[ 8:10] )
            # minute = int( version[ 10:12] )

            version_date = date( year, month, day )
            deltaTime = abs( today - version_date )
            deltaTime_days = deltaTime.days

            if ( self.upperDays < deltaTime_days ):
                dirName = self.backupDir + "/" + self.historyDir + "/" + version
                remove_a_folder( dirName )
                print "remove %s" %(dirName)
      
            pass




__doc__= """RSYNC_BACKUP  

Usage:  script.py [-r BACKUP_PLAN     ] 
                  [-b MAIN_BACKUP_DIR ]   
                  [-f FILE            ]
                  [-t DATETIME        ]
                  [--show_history_all ]  
                  [--show_version_all ]
                  [--preview          ]
                  [--preview_deleted  ]
                  [--recover          ]
                  [--recover_deleted  ]
                  [--print_deleted    ]
                  [--clean_up_versions]
                  [-hv]   

Options:
    -h --help                Show this screen
    -v --version             Show version.
    -r --read BACKUP_PLAN    Read in a backup plan, and process it. 
    -b --backup_dir MAIN_BACKUP_DIR  set backup folder
    -f --file FILE
    -t --datetime DATETIME 
    --show_all_history
    --show_version_all 
    --preview 
    --recover
    --print_deleted 
    --clean_up_versions
    
    example
    # process a backup plan
    ./script -r backup_plan

    # show all history versions (for all files)
    ./script -b ./backup --show_history_all

    # print all history versions for a given file 
    ./script -b ./backup -f temp.txt --show_version_all  
    
    # preview a previous version for a given file   
    ./script -b ./backup -f temp.txt -t 201901010005 --preview       

    # roll back to the previous version for a given file 
    ./script -b ./backup -f temp.txt -t 201901010005 --recover 

    # print all delelted files
    ./script -b ./backup --print_deleted
    
    # preview a deleted file   
    ./script -b ./backup -f xsel -t 201901032149 --preview_deleted 

    # recover a deleted file     
    ./script -b ./backup -f xsel -t 201901032149 --recover_deleted

"""

if __name__ == '__main__':
    
    rsync_tool = RSYNC()

    arguments = docopt(__doc__, version='RSYNC_BACKUP1.0' )
    # print("\n this is the arumgents from docopt:\n",  arguments )
    
    # process a backup plan
    if( arguments['--read'] ):
        rsync_tool.read_in_backup_plan( arguments['--read'] )
    

    # ./script -b ./backup --show_history_all
    if( arguments['--backup_dir'] and arguments['--show_history_all'] ):
        rsync_tool.set_main_backup_folder( arguments['--backup_dir'] )
        versions  = rsync_tool.get_history_versions()
        for v in versions: print v
    

    # print out date time info of versions for a given file     
    # ./script -b ./backup -f temp.txt --show_version_all   
    if( arguments['--backup_dir'] and\
        arguments['--file'] and\
        arguments['--show_version_all']  ):
        rsync_tool.set_main_backup_folder( arguments['--backup_dir'] )
        versions  = rsync_tool.get_file_versions( arguments['--file'] )
        for v in versions: print v


    # preview the previous version for a given file   
    # ./script -b ./backup -f temp.txt -t 201901010005 --preview       
    if( arguments['--backup_dir'] and\
        arguments['--file'] and\
        arguments['--datetime'] and\
        arguments['--preview']  ):
        
        rsync_tool.set_main_backup_folder( arguments['--backup_dir'] ) 
        
        rsync_tool.view_previous_version( arguments['--file'], \
                                          arguments['--datetime'] ) 

    # roll back to the previous version for a given file   
    # ./script -b ./backup -f temp.txt -t 201901010005 --recover       
    if( arguments['--backup_dir'] and\
        arguments['--file'] and\
        arguments['--datetime'] and\
        arguments['--recover']  ):
        
        rsync_tool.set_main_backup_folder( arguments['--backup_dir'] ) 
        
        rsync_tool.back_to_previous_version( arguments['--file'], \
                                          arguments['--datetime'] ) 
        

    # print all delelted files
    # ./script -b ./backup --print_deleted
    if( arguments['--backup_dir'] and\
        arguments['--print_deleted'] ):

        rsync_tool.set_main_backup_folder( arguments['--backup_dir'] )
        rsync_tool.show_all_deleted_files( ) 
         

    # preview a deleted file   
    # ./script -b ./backup -f xsel -t 201901032149 --preview_deleted 
    if( arguments['--backup_dir'] and\
        arguments['--file'] and\
        arguments['--datetime'] and\
        arguments['--preview_deleted']  ):  

        rsync_tool.set_main_backup_folder( arguments['--backup_dir'] ) 
        rsync_tool.preview_a_deleted_file( arguments['--file'],\
                                           arguments['--datetime'] )

    # recover a deleted file     
    # ./script -b ./backup -f xsel -t 201901032149 --recover_deleted        
    if( arguments['--backup_dir'] and\
        arguments['--file'] and\
        arguments['--datetime'] and\
        arguments['--recover_deleted']  ):
        
        rsync_tool.set_main_backup_folder( arguments['--backup_dir'] ) 
        
        rsync_tool.recover_a_deleted_file( arguments['--file'], \
                                          arguments['--datetime'] ) 
    

    # ------------------- demo ---------------------------------    
    if( 0 ): 
        rsync_tool.read_in_backup_plan( "backup_plan.txt" )

    if( 0 ):
        rsync_tool.set_main_backup_folder( "./backup")
        rsync_tool.add_source( "./dir" )
        rsync_tool.add_source( "temp.txt" )
        rsync_tool.process()
     
  
        versionsB = rsync_tool.get_file_versions( "temp.txt" )
        versionsB = rsync_tool.get_file_versions( "./dir/simple.txt" )
        print "----------------- get_file_versions"
        for v in versionsB: print v


        print "----------------- view_previous_version"
        rsync_tool.view_previous_version("temp.txt", "201901010005" )
        rsync_tool.view_previous_version("./dir/simple.txt", "201901022123" )

        print "----------------- back_to_previous_version"
        rsync_tool.back_to_previous_version("temp.txt", "201901010005" ) 
        rsync_tool.back_to_previous_version("temp.txt", "201901010011" ) 

        print "----------------- get_modified_files" 
        files1 = rsync_tool.get_modified_files()
        for f in files1: print f

        print "----------------- get_deleted_files" 
        files2 = rsync_tool.get_deleted_files()
        for f in files2: print f

        print "----------------- manage_history_version" 
        rsync_tool.manage_history_version()

        print "----------------- show_all_deleted_files" 
        rsync_tool.show_all_deleted_files()

        print "----------------- preview_a_deleted_file" 
        rsync_tool.preview_a_deleted_file( "dir/simple3.txt", "201901042334")
        rsync_tool.preview_a_deleted_file( "xxxx.123", "201901042300")


        rsync_tool.recover_a_deleted_file( "dir/simple3.txt", "201901042334")


  