#include <iomanip>
#include <iostream>
#include <vector>
using namespace std;
/*
 * Rvalue and lvalue.
 *
 * Simplified definition
 * lvalue = an object that occupies some identifiable location in "memory" 
 *         ( has * address )
 * Rvalue = any object that is not a lvalue.
 */

class demo_class
{

    public:
   
    int size;

    double* _arrary;
    
    demo_class( )
    {
	size = 5;

	_arrary = new double [ size ];

	for( int i = 0 ; i< size; i++ ) { _arrary[i] = 0 ; }
    }

    demo_class(  demo_class&& rhs ) // pass by rvalue
    { 

	size = rhs.size;

	_arrary = rhs._arrary;

	rhs._arrary = nullptr; 
	// we have to reset it.
	// since the when rhs object is destroyed( by deconstructor)
	// the _array will be gone.
    }

    ~demo_class() { delete _arrary; }

};

// create an object from demo_class, and return it.
demo_class create_demo_class() { demo_class obj; return obj; }


int sum( int x, int y) { return x+y; }

// a function take lvalue reference.
int square( int& x) { return x*x; }; 
// we cannot pass square( 5 )


// workaround 
int square( const int& x ) { return x*x; }
// add "const" such that it equal to a two steps:
// int temp = 5
// int& x = temp 
// so we can pass square( 5 );
// also square( iii ) works. 

int main( ) {


	//========================================== lvalue example
	int i; // i is a lvalue,
	       // since i's memory address can be found by &i

	i = 2; // Memory content can be modified to different value.

	demo_class obj1; // obj1 is a lvalue.




	//============================================ rvalue example
	int x = 2;  // 2 is a rvalue.
	x = (i+2);  // (i+2) is a rvalue.
	            // since we cannot get the (i+2)'s memory address
		    // no &(i+2), i is an integer. 

	demo_class obj2;

	obj2 = create_demo_class(); 
	// "create_demo_class()" is an rvalue
	// of user defined type. 
	
	int ii = sum(3,4);  
	// "sum(3,4)" is a rvalue 

	// ============================================ lvalue reference
	
	int iii;
	int & a_lvalue_ref = iii; 
	// iii is a lvalue
	// we use one '&' to do lvalue reference.
	// we cannot do lvalue reference to a rvalue.
	// int & ref = 5; is not valid. 
	// '5' is a rvalue.

	iii = 10;
	square( iii ); // it is OK, since iii is a lvalue, and then lvalue ref work.
    //  square( 40  ); // it is NOT OK for "int &x". 40 is a rvalue. 




	// ============================================ lvalue <--> rvalue.
	
	int i4 = 1;  // i4 is a lvalue.
	int x4 = i4; // now i4 is transformed to a rvalue in this expression.

	int v[3];
	*(v+2) = 4; // (v+2) is a rvalue. but when it is for an array index
		    //  we can get its address.
		    //  and deference *(v+2) is a lvalue.

	// not all the lvalue can be modified.
	const int i5 = 99; // lvalue i5 is not modifiable.  



}

