#include <iomanip>
#include <iostream>
#include <vector>
using namespace std;
/*
 *  move constructor
 */

// my class has both copy and move constructor.
class my_class 
{
    public:
    int size;
    double* _arrary;
    
    my_class( )  
    {
	size = 5;
	_arrary = new double [ size ];
	
	// costly deep copy.
	for( int i = 0 ; i< size; i++ ) { _arrary[i] = 0 ; }
    }

    // copy constructor.
    // usage: my_class obj1( obj2 );
    // it will call obj1.my_class( obj2 ); 
    my_class( const my_class & rhs ) 
    {
	size = rhs.size;
	_arrary = rhs._arrary;
    }
   
    // move constructor
    // usage: my_class obj1(  create_obj() )
    // it is common to overload the copy constructor by the
    my_class(  my_class&& rhs )  // no 'const' keyword
    { 
	size = rhs.size;
	_arrary = rhs._arrary;

	rhs._arrary = nullptr; 
	// we have to reset it.
	// since the when rhs object is destroyed( by deconstructor)
	// the _array will be gone.
    }

    // usage: obj1 = obj2
    operator=( my_class const & rhs); // lvalue reference. 

    // it is common to overload the copy operator by move operator.
    // usage: obj1 = create_object() 
    operator=( my_class && rhs);      // rvalue reference.

    ~my_class() { delete _arrary; }

};

// create an object from my class, and return it.
my_class create_obj() { 
    my_class obj; return obj; 
}


// dummy function, just put the my_class obj in.
// it takes r-value.
void func_A( my_class obj ) { };



//============= function overlad ========================

void printInt( int& i ) // 'int &'  take a 'lvalue', and we can change 'lvalue'
{ 
    cout <<"lvalue reference " << i << endl; 
}


void printInt( int&& i) // 'int &&' take a 'rvalue'
{   
    cout <<"rvalue reference " << i << endl; 
}
//======================================================


int main( ) {


    int  a = 5; // a is a lvalue( has memory address)     

    int& b = a; // b is a lvalue reference (int & )

    //int&& c;    // c is a rvalue reference (int && ).

    cout << "===========demo1: function overload ============" << endl;
    printInt( a ); // will call printInt( int&  i ) 

    printInt( 5 ); // will call printInt( int&& i ) 
		   // rvalue is a temp value, and will be destroyed. 

    cout << "===========demo2: move constructor ============" << endl;

    func_A( create_obj()  );
    // create_obj will return a 'my_class' obj 
    // which is a rvalue.
    // then it pass into the func_A's parameter
    // ie. my_class obj( create_obj() ) 
    // ps. create the obj, and pass argument into its constructor.
    // then, my_class(  my_class&& rhs ) constructor is invoked. 
    // we move the data from rhs to the current obj.


    // alternatively, we can do the following.

    my_class obj = create_obj();

    func_A( std::move( obj ) ); 
    
    // 'std::move()' will make the obj to a rvalue,
    // so we can call the move constructor.
    // and remember, obj._arrary will be nullptr after calling the move
    // constructor.
}

