#include <iomanip>
#include <iostream>
#include <vector>
using namespace std;
/*
 *  move constructor
 */

// my class has both copy and move constructor.
class my_class 
{
    public:
    int size;
    double* _arrary;
    
    my_class( )  //
    {
	size = 5;
	_arrary = new double [ size ];
	
	// costly deep copy.
	for( int i = 0 ; i< size; i++ ) { _arrary[i] = 0 ; }
    }

    // copy constructor.
    // usage: my_class obj1( obj2 );
    // it will call obj1.my_class( obj2 ); 
    my_class( const my_class & rhs ) 
    {
	size = rhs.size;
	_arrary = rhs._arrary;
    }
   
    // move constructor
    // usage: my_class obj1(  create_obj() )
    // it is common to overload the copy constructor by the
    my_class(  my_class&& rhs )  // no 'const' keyword
    { 
	size = rhs.size;
	_arrary = rhs._arrary;

	rhs._arrary = nullptr; 
	// we have to reset it.
	// since the when rhs object is destroyed( by deconstructor)
	// the _array will be gone.
    }

    // usage: obj1 = obj2
    operator=( my_class const & rhs); // lvalue reference. 

    // it is common to overload the copy operator by move operator.
    // usage: obj1 = create_object() 
    operator=( my_class && rhs);      // rvalue reference.

    ~my_class() { delete _arrary; }

};

// create an object from my class, and return it.
my_class create_obj() { 
    my_class obj; return obj; 
}


// dummy function, just put the my_class obj in.
// it takes r-value.
void func_A( my_class obj ) { };



template< typename T > 
void relay( T&& arg ) {

    // we pass the arg from 'relay' to 'func_A'
    // ( it call argument forwarding )
    // arg can be either lvalue or rvalue.
    // we wish forward the same type rvalue-->rvalue, lvalue-->lvalue.
    // here T&& is a unverisal template. 
 
   func_A( std::forward<T> arg ); 
}


/*
    reference collasping rules (c++11)

    T&  &   ==> T&  ( T reference to a reference will deduced to a T reference )
    T&  &&  ==> T&
    T&& &   ==> T&
    T&& &&  ==> T&&

    The compiler's rule.
*/

/*

    remove_reference<int&>::type i;  // it is equal to  int i;

    remove_reference<int>::type i;   // int i;
    // since 'int' has no reference to remove,
    // it will be equal to 'int'
 
*/

/*
    call relay( 9 ), and 9 is a rvalue 
    T   => int &&
    T&& => int && && ==> int &&
    arg has a type of rvalue.

    call ralay( x ), and int x = 9; x is a lvalue.
    T   => int &
    T&& => int & && ==>  int & 
    arg has a type of lvalue.
*/

// forward function: what is it doing?
// it will cast the arg back to the T&& and pass to func_A()
/*
 * template<class T>
 * T&& forward( typename remove<T>::type& arg) {
 *  return static_cast<T&&>(arg);
 * }
 */ 

int main( ) {


    func_A( create_obj()  );
    // create_obj will return a 'my_class' obj 
    // which is a rvalue.
    // then it pass into the func_A's parameter
    // ie. my_class obj( create_obj() ) 
    // ps. create the obj, and pass argument into its constructor.
    // then, my_class(  my_class&& rhs ) constructor is invoked. 
    // we move the data from rhs to the current obj.


    // alternatively, we can do the following.

    my_class obj = create_obj();

    func_A( std::move( obj ) ); 
    
    // 'std::move()' will make the obj to a rvalue,
    // so we can call the move constructor.
    // and remember, obj._arrary will be nullptr after calling the move
    // constructor.
}

