#include <iomanip>
#include <iostream>
#include <vector>
using namespace std;


/*
 user defined literals (constant)
 the build-in type
 integer        ---  45
 floating point ---  4.5
 character      --- 'z'
 string         --- "xination"

 45;   // string literal
 45u;  // u suffix means unsigned ==> unsigned int.
 45l;  // l suffix means long     ==> long int.

 in c++11, we can define our own suffix.
*/

// user defined literals
// SI unit meter, kilogram
// the input arg for float is 'long double' 
// use 'constexpr' to do the unit transfromation at the compiler time.
constexpr double operator"" _cm( long double x )  { return x * 0.1; }
constexpr double operator"" _m(  long double x )  { return x ; }
constexpr double operator"" _um( long double x )  { return x * 1E-6; }
constexpr double operator"" _kg( long double x )  { return x ; }
constexpr double operator"" _g ( long double x )  { return x * 1E-3 ; }
/*
ReturnType operator "" _a(unsigned long long int);   // Literal operator for user-defined INTEGRAL literal  
ReturnType operator "" _b(long double);              // Literal operator for user-defined FLOATING literal  
ReturnType operator "" _c(char);                     // Literal operator for user-defined CHARACTER literal  
ReturnType operator "" _d(wchar_t);                  // Literal operator for user-defined CHARACTER literal  
ReturnType operator "" _e(char16_t);                 // Literal operator for user-defined CHARACTER literal  
ReturnType operator "" _f(char32_t);                 // Literal operator for user-defined CHARACTER literal  
ReturnType operator "" _g(const     char*, size_t);  // Literal operator for user-defined STRING literal  
ReturnType operator "" _h(const  wchar_t*, size_t);  // Literal operator for user-defined STRING literal  
ReturnType operator "" _i(const char16_t*, size_t);  // Literal operator for user-defined STRING literal  
ReturnType operator "" _g(const char32_t*, size_t);  // Literal operator for user-defined STRING literal  
ReturnType operator "" _r(const char*);              // Raw literal operator  
template<char...> ReturnType operator "" _t();       // Literal operator template 
*/


int operator"" _bin( const char* str, size_t l )
{
    int ret = 0;

    for( int i=0; i<l; i++) {
    
	ret = ( ret << 1 ); 
	// '<<' is to left shift by 1
	// 0001 ==> 0010
	
	if( str[i] == '1' ) 
	{
	    ret += 1; 
	}

    }

    return ret;
}

int main( ) 
{
    cout << "========== demo 1 ================" << endl;

    double x = 1.0_um; 
    double y = 2.0_um; 
    double z = 3.0_um; 

    double w = 20.0_g;

    cout << "density = " << w/x/y/z << " [kg/m^3] " << endl;
   

    cout << "========== demo 2 ================" << endl;

    cout << "1"_bin << endl;
    cout << "10"_bin << endl;
    cout << "11"_bin << endl;
    cout << "111"_bin << endl;
}

