#include <iomanip>
#include <iostream>
#include <string>
#include <memory> // to use smart pointer.
using namespace std;


/*
    unique pointer
    * light weight point
    * exclusive owership ( cannot be copied, no ptr1=ptr2 is allowed.)
     
*/

// a class to demo the smart pointer.
class Dog 
{
    private:
	string _name;
	//shared_ptr<Dog> my_Friend; ( this will cause Cyclic reference) 
    public:
    Dog(){ _name ="nameless";}      // default constructor. Dog dog1():
    Dog( string name )
    { 
	_name = name; 
	cout << "Dog " << _name << " is created" << endl;
    }


    Dog( const Dog&  rhs);          
    Dog& operator=(const Dog& rhs);  
    ~Dog(){
	cout << "Dog " << _name << " is destoryed" << endl;
    };

    Dog( Dog&& );            
    Dog& operator=( Dog&& ); 

    void bark() 
    {
	    cout << "Dog " << _name << " barks." << endl;
    }


	
};


void func_demo() 
{

    unique_ptr<Dog> ptr1 ( new Dog("Gunner") );

    // check the ptr1 is null or not.
    
    if( ptr1 ) { cout << "ptr1 is not empty" << endl; }
    else { cout << "ptr1 is empty" << endl; }

    Dog* d = ptr1.release();
    // relese() will release the owership of the unique_ptr
    // and it return the raw pointer.
    // it is similar to the shared_ptr's get() function.


    ptr1.reset( new Dog("Smokey") );
    // reset it another Dog 
    // note: ptr1.reset() ==> set it to nullptr
}

void test_dog ( unique_ptr<Dog> p ) { p->bark(); }

unique_ptr<Dog> getDog() {
    unique_ptr<Dog> p( new Dog("Tom") );
    
    return p; 
    // it will automatically use move semantics to return the value.
    // p is not the owner of "Tom" anymore.
}

void func_demo2() 
{
    unique_ptr<Dog> ptr1 ( new Dog("Gunner") );
    unique_ptr<Dog> ptr2 ( new Dog("Smokey") );

    // use std::move to move the data in ptr1 "Gunner"
    // to ptr2 "Smokey", at these steps are:
    // 1. "Smokey" is destoryed.
    // 2. ptr1 is empty
    // 3. ptr2 owns "Gunner"
    ptr2 = std::move( ptr1 );
    
    test_dog( std::move(ptr2) ); 
    // when passing into a function, we need to use move
    // since
    // unique_ptr(const unique_ptr&) = delete;

    unique_ptr<Dog> ptr3 = getDog();
}

void func_demo3 () 
{
    // unique_ptr is designed for array.
    // we don't need to write our own deleter, like shared_pointer.
    // shared_ptr<Dog> p( new Dog[3], 
    // [] (Dog* p) { delete [] p } );

    unique_ptr<Dog [] > dogs( new Dog[3] );

}


int main() {

    func_demo();

    func_demo2();
}
