#include <boost/variant.hpp>
#include <vector>
#include <iostream>
#include <string>

using namespace std; 

/*
    variant<T1, T2> class 
    can let you store T1 "or" T2 value in an obj.
    
    variant<string, int> 
    ex. myVariant = "123";
    ex. myVariant = 123;

    == working on variant
    boost::get<T>
    boost::apply_visitor()
    boost::static_visitor<>   
    
*/

 
class Vistor_DoubleTheValue : public boost::static_visitor<> {
public:
   void operator() (int& i) const {  
       i += i;
   }
   void operator() (string& str) const { 
       str += str;
   }
};

int main( ){
    

    // demo1 
    cout << "\ndemo1:" << endl;
    boost::variant<int, string> myVariant;

    string varString = "123";
    int    varInt = 123;

    myVariant = varString;
    cout << myVariant << endl;

    myVariant = varInt;
    cout << myVariant << endl;

    //================================
    // demo2 
    cout << "\ndemo2:" << endl;
    // to do arithmetic of variant object, 
    // we use the following fashion.


    myVariant = boost::get<int>( myVariant )* 2;
    // reminder, in the previous step, myVariant = varInt
    // we cannot do get<int>( string ),
    // boost will give you an exception.
    cout << myVariant << endl; 
    // we get an output: 246

    myVariant = varString;
    myVariant = boost::get<string>( myVariant ) + "~~Hello" ;
    cout << myVariant << endl; 
    // we get an output "123~~Hello"



    //================================
    // demo3 
    cout << "\ndemo3:" << endl;
    // a variant object is never empty
    // it will be the first type, ie. T1 type.
    // T2 type value is not defined yet.
    boost::variant<int, string> testVariant;
    cout << boost::get<int>( testVariant ) << endl;
    // we get an output 0



    //================================
    // demo4 
    cout << "\ndemo4:" << endl;
    // work on variant no matter which type it is now.

    //Using visitor
    vector< boost::variant<int, string> > containerVector;
    containerVector.push_back("good");
    containerVector.push_back(25);
    containerVector.push_back("bad");
    for (auto x : containerVector) {

        boost::apply_visitor( Vistor_DoubleTheValue(), x );
        cout << x << endl;
    }
    /*  we get:
        goodgood 
        50
        badbad
    */

    return 0;
}