#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>
#include <string>
#include <iostream>
using namespace std;
using boost::lexical_cast;

/*
    ref: https://theboostcpplibraries.com/boost.tokenizer
    summary:

    1. shorthand type
    tokenizer type =  boost::tokenizer< boost::char_separator<char> >


    2. seperator object ( if we don't want to seperate by spaces )
    char_separator<char> sep{ "@" }


    3. create a tokenizer to store the tokenized string.
    tokenizer tok_container{ varString, sep };



*/

int main( ){

    // demo1 
    cout << "\n=======================demo1:" << endl;

    // define a shorted notation 
    // we use tokenizer<T>, where T = char_separator<char> 
    // char_separator<char> denotes the separation by "space(s)"
    typedef boost::tokenizer< boost::char_separator<char> > tokenizer;
    
    string varString = "How are you doing? ";
    
    tokenizer tok_container{ varString };
    // note { varString } is to intialize tok_container with a list.
    // it will invoke the constructor, where it takes:
    // (const initializer_list<string> &v)

    for ( auto itr : tok_container ){ cout << itr << '\n';  }
    
    //  boost::char_separator interprets spaces and punctuation marks as
    //  separators by default
    //  we get the following:
    //  How
    //  are
    //  you
    //  doing
    //  ?
    


    // demo2 
    // we can of cousre use another deliminator, such as @
    cout << "\n=======================demo2:" << endl;
    
    varString = "xination@mail.com";
    
    boost::char_separator<char> sep2 { "@" };

    tokenizer tok_container2{ varString, sep2 };
    for ( auto itr : tok_container2 ){ cout << itr << '\n';  }



    // demo3 
    // we can add a display deliminator.
    cout << "\n=======================demo3:" << endl;
    
    varString = "xination@mail.com.";
    boost::char_separator<char> sep3 { "@", "." };
    // "@" is the first  deliminator, and the @ will be all gone.
    // "." is the second deliminator, and it will be displayed.
    
    // we get the following.
    // xination
    // mail
    // .
    // com
    //.

     
    tokenizer tok_container3{ varString, sep3 };
    for ( auto itr : tok_container3 ){ cout << itr << '\n';  }    


    cout << "\n=======================demo4:" << endl;
  
    varString = "100.1 200. 300. 5.5555";
    boost::char_separator<char> sep4{" "};
    tokenizer tok_container4{ varString, sep4 };
    
    
    for ( auto itr : tok_container4 ){ 
        // note: itr is a std::string type.
        cout << lexical_cast<double>( itr )  << '\n';  
    }

    return 0;
}