#include <iostream> // for cout
#include <bitset>
#include <iomanip>  // to show hex

using namespace std;


/*
    http://www.learncpp.com/cpp-tutorial/38-bitwise-operators/


    operator	asm equivalent	description
    _____________________________________________________________
    &	        AND	            Bitwise AND
    |	        OR	            Bitwise inclusive OR
    ^	        XOR	            Bitwise exclusive OR
    ~	        NOT	            Unary complement (bit inversion)
    <<	        SHL	            Shift bits left
    >>	        SHR	            Shift bits right



    bytes have 8 bits : _ _ _ _ _ _ _ _
    looks like          1 0 0 0 0 0 0 0  (128)
                        0 1 0 0 0 0 0 0  ( 64)
                        0 0 1 0 0 0 0 0  ( 32)
                        0 0 0 1 0 0 0 0  ( 16)
                        0 0 0 0 1 0 0 0  (  8)
                        0 0 0 0 0 1 0 0  (  4)
                        0 0 0 0 0 0 1 0  (  2)
                        0 0 0 0 0 0 0 1  (  1)



    As with the arithmetic assignment operators,
    C++ provides bitwise assignment operators in order to
    facilitate easy modification of variables.

    Operator 	            Symbol 	Form 	    Operation
    _____________________________________________________________

    Left shift assignment 	<<= 	x <<= y 	Shift x left by y bits
    Right shift assignment 	>>= 	x >>= y 	Shift x right by y bits
    Bitwise OR assignment 	|= 	    x |= y 	    Assign x | y to x
    Bitwise AND assignment 	&= 	    x &= y 	    Assign x & y to x
    Bitwise XOR assignment 	^= 	    x ^= y 	    Assign x ^ y to x

    For example, instead of writing x = x << 1;, you can write x <<= 1;


*/





int main() {

    //
    // basic information about bit
    //


    unsigned short num;
    // short has two bytes ( each byte has 8 bits)
    // in the following  example, we just show one byte ( 8 bits )
    num = 0; cout << bitset<8>( num ) << endl;  // we get 0000 0000
    num = 1; cout << bitset<8>( num ) << endl;  // we get 0000 0001
    num = 2; cout << bitset<8>( num ) << endl;  // we get 0000 0010
    num = 3; cout << bitset<8>( num ) << endl;  // we get 0000 0011
    num = 4; cout << bitset<8>( num ) << endl;  // we get 0000 0100
    num = 8; cout << bitset<8>( num ) << endl;  // we get 0000 1000
    num =16; cout << bitset<8>( num ) << endl;  // we get 0001 0000
    num =32; cout << bitset<8>( num ) << endl;  // we get 0010 0000
    num =64; cout << bitset<8>( num ) << endl;  // we get 0100 0000
    num =128; cout << bitset<8>( num ) << endl; // we get 1000 0000
    num =255; cout << bitset<8>( num ) << endl; // we get 1111 1111

    num =127; cout << bitset<8>( num ) << endl; // we get 0111 1111 ( common mask )

    // hex-decimal
    cout << showbase      // show the 0x prefix
         << internal      // fill between the prefix and the number
         << setfill('0'); // fill with 0s

    num = 10; cout << std::hex << num  << std::dec << endl; // we get 0xa
    num = 11; cout << std::hex << num  << std::dec << endl; // we get 0xb
    num = 12; cout << std::hex << num  << std::dec << endl; // we get 0xc
    num = 13; cout << std::hex << num  << std::dec << endl; // we get 0xd
    num = 14; cout << std::hex << num  << std::dec << endl; // we get 0xe
    num = 15; cout << std::hex << num  << std::dec << endl; // we get 0xf


    //
    //  bitwise and:  &
    //
    if( 0 ) {

        /*
        bit a 	bit b 	a & b (a AND b)
        0 	    0 	    0
        0 	    1 	    0
        1 	    0 	    0
        1 	    1 	    1
        */

        int num_a, num_b;
        num_a =128; cout <<" num_a = " <<  bitset<8>( num_a ) << endl; // we get 1000 0000
        num_b =255; cout <<" num_b = " <<  bitset<8>( num_b ) << endl; // we get 1111 1111

        cout << " num_b & num_a = " << bitset<8>( num_b & num_a ) << endl;

    }

    //
    // left-shift  <<
    // right-shift >>
    //
    if( 0 ) {


        int num = 1;
        // left-shift will move whole bit array to the left, and fill 0 to the most right.
        cout << "num      we get " << bitset<8>( num )      << endl; // we get 0000 0001
        cout << "num << 1 we get " << bitset<8>( num << 1 ) << endl; // we get 0000 0010
        cout << "num << 2 we get " << bitset<8>( num << 2 ) << endl; // we get 0000 0100
        cout << "num << 3 we get " << bitset<8>( num << 3 ) << endl; // we get 0000 1000

        cout << "\n\n" << endl;



        num = 128;
        // right-shift will move whole bit array to the right, and fill 0 to the most left.
        cout << "num      we get " << bitset<8>( num )      << endl; // we get 1000 0000
        cout << "num >> 1 we get " << bitset<8>( num >> 1 ) << endl; // we get 0100 0000
        cout << "num >> 2 we get " << bitset<8>( num >> 2 ) << endl; // we get 0010 0000

    }

    //
    // bitwise xor and or:  
    //
    if( 1 ) {

		//note: unsigned long has 4 bytes = 16 bits.
        unsigned long n1 = 10;       
        unsigned long n2 = 12345678;


        unsigned long n3;
        unsigned long n4;
		n3 = n2 ^ n1;  
		n4 = n2 | n1;  
        
        cout << "n1 = "<< setw(12) << bitset<12>( n1 ) << endl; 
        cout << "n2 = "<< setw(12) << bitset<12>( n2 ) << endl; 
		cout << "n3 = "<< setw(12) << bitset<12>( n3 ) << endl; 
		cout << "n4 = "<< setw(12) << bitset<12>( n4 ) << endl; 
		// n1 = 0000 0000 1010
		// n2 = 0001 0100 1110
		// n3 = 0001 0100 0100  ( by ^, xor eclusive or ) 
		// --- (0,0) = (1,1) = 0, while ( 1,0) = (0,1) = 1

		// n1 = 0000 0000 1010
		// n2 = 0001 0100 1110
		// n4 = 0001 0100 1110  ( by |, or ) 
		// --- (0,0) = 0, (1,0) = (0,1) = (1,1) = 1
	}


}

