#include <iostream>
#include <vector>
#include <algorithm> // to use sort

using namespace std; // STL library is in the std namespace.

/*
    introduction:
    algorithm only works on the iterator, and don't directly work
    on the various sorts of data containers. 
    it is much re-usable. 

    many compliers are tuned to optimize the STL.
*/



int main() {

    vector<int> vec;

    vec.push_back( 4 );
    vec.push_back( 1 );
    vec.push_back( 8 );

    // [ begin, end ), not include the end.
    vector<int>::iterator itr1 = vec.begin();
    vector<int>::iterator itr2 = vec.end();
    
    // vec.end() points to the one block after the last value. 
    // we have an error when doing 
    // cout << *vec.end() << endl;

    //use the iterator class from meta class vector, which is set to <int>

    // looping.
    for( vector<int>::iterator itr = itr1; itr != itr2; ++itr ) 
    {
	cout << *itr << " ";
	// behave like a pointer ( we can de-reference it to get value )
    } 
    cout << "\n";


    // algorithm
   
    sort( itr1, itr2 ); 
    // sort work on iterator !!!!
    // after sorting vec = { 1, 4, 8 }
    for( auto itr = itr1; itr != itr2; ++itr ) 
    {
	cout << *itr << " ";
    }
    cout << "\n";
}
