#include <iostream>
#include <vector>
#include <deque>
#include <list>
#include <set> // with set and multiset class
#include <map> // with map and multimap class
#include <unordered_set> // with unordered_set/multiset
#include <unordered_map> // with unordered_map/multimap
#include <algorithm> 
#include <numeric>      // some numeric algorithm
#include <functional>   // functor.

using namespace std; // STL library is in the std namespace.

/*
    
     
    //=================================================================
    // property of associative containder
    // set, multiset,
    // map, multimap.
    always sorted ...implemented in binary tree, 
    default criteria is '<'
    No push_front() and push_back()




*/



int main() {

    //=================================================================
    // property of set:
    // No duplicate items
    // insert is O( log(n) )
    // Fast : search is O( log(n) )
    //        sequence container takes linear time O(n) to search.
    // No random access
    // value of element cannot be modified.
    set<int> myset;

    myset.insert( 3 );
    myset.insert( 1 );
    myset.insert( 7 ); // myset: { 1, 3, 7}, note: it automatically sorted.
    
    set<int>::iterator itr;
    itr = myset.find(7);  // O( log(n) )
    
    cout << " TEST itr" << endl;
    cout << *itr << endl; 
    // note: sequence containers don't have find method.
    // we have to use find function.
    // ex. for a list object
    // itr = find( mylist.begin(), mylist.end(), 5 );






    // insert() method also return a pair of values.
    // the first  is the iterator of to the insert position.
    // the second is the bool to indicate the insertion successful or not.
    pair< set<int>::iterator, bool> ret;
    ret = myset.insert(3);
    // since 3 has been in our set, it would not be implemented.
    if( ret.second == false ) { cout << " myset.insert(3) would not work \n"; }
    for( auto i : myset ) { cout << i << " "; } cout << "\n=========\n";


    // if we provide a good hint for insertion can reduce the insertion time.
    // from O( log(n) ) to O( 1 ) 
    itr = myset.find( 7 ); 
    myset.insert( itr, 9 ); 
    // the first par is for the hint of location.
    // the second par is for value.
    cout << "TEST insert to a set:\n";
    for( auto i : myset ) { cout << i << " "; } cout << "\n=========\n";





    // erase
    myset.erase( 9 );    // by value ( not for sequence container )
    myset.erase( itr );  // by iterator.
    cout << "TEST erase:\n";
    for( auto i : myset ) { cout << i << " "; } cout << "\n=========\n";





    //=================================================================
    // property of multiset:
    // allow duplicate items
    // insert is O( log(n) )
    // Fast : search is O( log(n) )
    //        sequence container takes linear time O(n) to search.
    // Slow : traversing is slow compare to vector and deque.
    // No random access
    // value of element cannot be modified. ( no *itr = 100 )
    // insertion into a multiset always successful.

  






    //=================================================================

    // 'map' and 'multimap' sort the value according to 'key' not 'value'
    // they have ( key, value ) pair
    // but 'set' and 'multiset' just use value as their key.


    // property of map:
    // no duplicate 'key'
    // keys cannot be modified. 
    // type of *itr = pair< type for key, type for value >
    
    map< char, int> myMap;
    
    // map object take a pair object as its input.
    // or use make_pair
    myMap.insert( pair<char, int>( 'a', 100 ) );
    myMap.insert( make_pair( 'b', 200 ) );

    auto map_itr = myMap.begin(); // map<char, int>::iterator itr
    myMap.insert( map_itr, make_pair( 'c', 300) );  // map_itr is a hint of location.

    // we can use [ key ] to modify the value for a given key.
    myMap['c'] = -300; 




    // find by key
    map_itr = myMap.find('c');
    cout << "\nmap find test \n";
    cout << "first = "    << (*map_itr).first 
	 << ", second = " << (*map_itr).second << endl;

    // or we can use at( key )
    cout << "myMap.at('c') = " << myMap.at('c') << endl; 
	
    cout << "\nprint all the key-value pair in a map \n";
    for( auto i: myMap ) {
	cout << "first = "    << i.first 
	     << ", second = " << i.second << endl; }


    //=================================================================
    // property of multimap:
    // allow duplicate 'key'
    // keys cannot be modified. 
    // type of *itr = pair< type for key, type for value >
    multimap<char, int> myMultiMap;

    // no following.
    // (*map_itr).first = 'x' 


}
