#include <iostream>
#include "header.h"
#include <string>
#include <vector>
using namespace std;

/* summary

    typename or class are all valid keyword.

    1. function template 
    template <typename T>
    T Max(T const& a, T const& b)

    
    2. class template
    template <typename T>
    class mypair { ... };

    template <typename T>
    T mypair<T>::getmax () {....} // for the implementation.


    3. variadic template; indefinte number of arguments (varia-dic)
    template<typename First, typename... Rest>
    void myfunc( First        first_arg,
                 Rest...      rest_args ) { 
        myfunc ( first_arg );
        myfunc ( rest_args... );
    }
    template<typename T>
    void myfunc(   T  first_arg   ) {...}     

*/



// example for function
// syntax:
// template <class type>  ret-type   func-name(parameter list)
template <typename T>
inline T Max(T const& a, T const& b)
{
    return a < b ? b : a;
    // to find the max between a and b
    // if (a < b) { return b } else { return a;}
}
// here inline is optional, but recommanded.
// it will be necessary when we deal with #include







// example for class
// syntax:
// template <typname type> class class-name {....}
// note   <typename type> = <class type>, some people use the latter one.
template <typename T>
class mypair {

    T a, b; // a and b variables have type 'T'


  public:

    mypair (T first, T second) {a=first; b=second;} // constructor

    T getmax ();  // for mehtod 'getmax', the return type is 'T'


};


// the 'getmax' method is defined here.
template <typename T>
T   mypair<T>::getmax ()
{
  T retval;
  retval = a>b? a : b;
  return retval;
}



// C++11 feature: indefinte number of arguments
// similar to python:
/*
    def info_show1( *argv ):
        for item in argv:
        print( item )
*/
template<typename... Args>
inline void myprintf(const char *s, Args... args )
{
    printf(s, args...);
    //  the args... is unpacked here.
    //  termilogy: variadic template. 
}



//A general technique
/*
    function overload and
    function recursion.
    in C++, there no good way to loop through the packed arguments.
    So we have to use this trick.
*/


template<typename T>
inline void myfunc
(   T  first_arg   )
{

    // application: TFile or fstream
    cout << " I am open the file = " << first_arg << endl;

}



template<typename First, typename... Rest>
inline void myfunc
( First        first_arg,
  Rest...      rest_args )
{
    myfunc ( first_arg );    // the myfunc with single parameter is called.

    myfunc ( rest_args... ); // this one will call the 'myfunc' with mutilple paramters,
                             // total arg number will be decrease by 1.
                             // until it has only one argument left.
}





//
// class template:
// we have to specific the type for T ( unlike funciton template )
// ie. demo_class<int> obj
// ( I remember it relates to constructor )
template<typename T>
class demo_class {
    T arr[100];
    int idx;
public:
    demo_class(): idx(0) {}
    void push( T x ) { arr[idx] = x; idx++; }
    T Get( int i ) const  { return arr[i]; }
};




int main() {

    if( 0) // function template demo.
    {
        int i = 39;
        int j = 20;
        cout << "test1 --  Max(i, j): " << Max(i, j) << endl;
        
    	// note:
    	// Max funciton is defined as
    	// template< typename T>
    	// inline T Max( T const & a, T const & b )
    	// we don't need to specify the type as the above case.
    	// the data type can also be inferred from the parameters.
    	// Indeed, we can call Max<int>( i, j) 

        float f1 = 40.1;
        float f2 = 20.2;
        cout << "test2 --  Max(f1, f2): " << Max(f1, f2) << endl;

        string s1 = "Hello";
        string s2 = "World";
        cout << "test3 --  Max(s1, s2): " << Max(s1, s2) << endl;

    }





    if (0) // class template demo
    {

        mypair<int> myobject (100, 75);
        /*
            when declare an obj from a template class,
            we need the class name and type.
            one very common is vector<int> my_array;
        */

        cout << myobject.getmax();
    }





    if( 0 ) // C++11 feature: indefinte number of arguments
    {

        myprintf("%s %s %d %s", // this part goes to const char* s
                 "xination", " is ", 32, " years old." // this part is packed into 'args'
                 );

    }



    if ( 1 ) // demo to create function with idefinite argements
             // then we subsequently deal with these arguments.
    {
        myfunc( "run01", "run02", "run03");
    }

    return 0;
}




