#include <iostream>
#include "header.h"
using namespace std;


/*	C++11 feature: indefinite number of arguments
	
	as recursion programming has two part 
	(1) base case -- or call escape part
	(2) recursive case. 

	variadic uses the similar technique to unpack arguments.
	suppose we pass func( int, double, char, short )

	the compiler actually generate:
	func( int ), func( double, char, short )
	
	and 
	func( double), func( char, short)

	and func( char ), func (short )

	for us to overload them. 
	---------------------------------------------------
	
	myfunc( int, double, char, short )
    when it pass to 
	template< typename First, typename ... Rest >
	void myfunc( First first_arg, Rest ... rest_args )

	the argument is grouped:
	the typename "First"  = int
	the typename "Rest" = ( double, char, short )

	then first_arg is pass to escape function.
	while the "rest_args..." is passed to the the same function (recursive)
	
	so the next round:
	the typename "First"  = double
	the typename "Rest" = ( char, short )

	so the next round:
	the typename "First"  = char
	the typename "Rest" = ( short ) => will call one argument base case. 
*/




/*
 * Step 1: we create the function that takes ONLY one argument.
 * In our case, it is myfunct( T one_arg )
 * escape case.
*/
template<typename T>
inline void myfunc (   T  one_arg   )
{

    // application: TFile or fstream
    cout << " I am open the file = " << one_arg << endl;

}


/*
    Step 2: we overlap the one-argument function 
    by a function take two different types of parameters.
    myfunc ( First first_arg, Rest... rest_args )
    the First first_arg is passing to the myfunc( T one_arg )
    but the Rest... rest_args are passing the same function itself.

    

    template<typename First, typename... Rest>
    typename    ==> a type for single object
    typename... ==> a type for multiple objects.
   
    for example, if Rest = TString
    then Rest... is a list of TString objects. 

    myfunc( rest_args... ), means passing a list of objects with "Rest" type.
    it will call myfunc( first_arg, rest_args ) again.

    the second myfunc always has this framework. 
    We just need to work on the single one.
*/




template<typename First, typename... Rest>
inline void myfunc ( First  first_arg,  Rest...  rest_args )
{
    myfunc ( first_arg );    // the myfunc with single parameter is called.

    myfunc ( rest_args... ); // this one will call the 'myfunc' with mutilple paramters,
                             // total arg number will be decrease by 1.
                             // until it has only one argument left.
}






int main() {

    myfunc( "run01", "run02", "run03");
    
    return 0;
}




