#include <iostream>
#include <algorithm>  // for transform
#include <vector>
using namespace std;

/**! summary
    
    functor is an object from a class where its operator () is overload,
    and so it likes a function.

    functor's advantage is that it has internal state (it is an object)

    we normally use std::transform( itr_input_first, 
                                    itr_input_last, 
                                    itr_output, 
                                    operation  )

*/

// A Functor 
class increment { 
private: 
    int fNum; 
public: 
    increment(int n) : fNum(n) {  } 
  
    // This operator overloading enables calling 
    // operator function () on objects of increment 
    // ie. obj.( arr_num ) ==> will call this method
    int operator () (int arr_num) const 
    { 
        return fNum + arr_num; 
    } 
}; 
 


int op_increase (int i) { return i+10; }

int main( ){


    // demo1:: functor 
    cout << "\ndemo1:" << endl;

    int arrN = 5;
    int arr[] = {1, 2, 3, 4, 5}; 

    
    // 1. instance an obj, increment obj( 10 ), 
    //    and taking 10 to its constructor 
    // 2. apply obj(arr[i]) => call obj.operator( arr[i] )
    //    return 10+arr[i]
    transform( arr, arr+arrN, arr, increment(10) ); 
    
    for (int i=0; i<arrN; i++){ cout << arr[i] << endl; }
    // we get 11, 12, 13, 14, 15


    // demo2: quick demo how to use std::transform
    cout << "\ndemo2:" << endl;

    vector<int> v1 = {1,2,3,4,5};
    vector<int> v2;
    v2.resize( v1.size() );

    transform (v1.begin(), v1.end(), v2.begin(), op_increase);
    
    for( auto itr: v2 ){ cout << itr << endl; }
    // we get 11, 12, 13, 14, 15    

    return 0;
}