#include <iostream> // for cout
#include <fstream>
#include <cstdlib>  // for exit.
void create_binary_file();
void read_binary_file();
using namespace std;




/*
|
|  a common skill to get the size of file.
|
|   STEP 1:
|     seekg (0, ios::end);
|     // move the position to the end of "infile", offset = 0,
|	  // it means offset to the ios::end 
|	  // we can also do seekg( ios::end), jump directly to the end.
|
|
|
|   STEP 2:
|    tellg() will return the position reading, and since it is the end of the file,
|    and we will get the size of file.
|
|    unsigned int filesize = infile.tellg();
|    hence we set up the "filesize" variable to retrieve the files.
|
|   STEP 3:
|    we reset our position at the beginning of the file.
|    infile.seekg (0, ios::beg);
|	  // we can also do seekg( ios::beg), jump directly to the beginning.
|   
|    ref http://www.cplusplus.com/doc/tutorial/files/
|   
*/



int main() {

    create_binary_file(); // create the data.dat file.

    fstream in;    in.open( "data.dat", ios::in | ios::binary);

    // warming information.
    if( in.fail() ) { cout << " having problem to open the file" << endl; exit( EXIT_SUCCESS); }


    //!------the key part-----

    int filesize;

    in.seekg(0, ios::end);  // use seek get method to jump to the end.

    filesize = in.tellg();  // tellg() returns the position reading in unit of byte. 

    in.seekg(0, ios::beg);  // use seek get method to jump to the beginning.

	// note if we reaches the end, the state good() = false.
	// we should reset by in.clear(); 

    //!-----------------------


    cout << "we read in " << filesize << " bytes" << endl;

    read_binary_file(); // to read out what inside the file.

}

//
// create a binary file for test
//
void create_binary_file() {

	fstream out;
	out.open("data.dat" , ios::out|ios::binary);

    // Here is also a reminding of how to write binary format.

    char   x_char   = 'z';
    short  x_short  = 2;
    int    x_int    = 3;
    float  x_float  = 4;
    double x_double = 5;

    char   y_char[2] = {'a','b'};
    int    y_int[2]  = {1, 2};
    float  y_float[2]= {0.5, 0.6};

    struct info_struct {
        int    z_int;
        float  z_float;
        double z_double[2];
    };

    info_struct info;

    info.z_int   = 9;
    info.z_float = 0.9;
    info.z_double[0] = 0.9;
    info.z_double[1] = 9.9;

    int bytes_counter = 0;

    out.write( &x_char, sizeof(x_char) );                               bytes_counter += sizeof( x_char );
    out.write( reinterpret_cast<char*>( &x_short ),  sizeof(x_short) ); bytes_counter += sizeof( x_short );
    out.write( reinterpret_cast<char*>( &x_int ),    sizeof(x_int) );   bytes_counter += sizeof( x_int  );
    out.write( reinterpret_cast<char*>( &x_float ),  sizeof(x_float) ); bytes_counter += sizeof( x_float );
    out.write( reinterpret_cast<char*>( &x_double ), sizeof(x_double) );bytes_counter += sizeof( x_double);

    out.write( reinterpret_cast<char*>( y_char),  sizeof( y_char) );    bytes_counter += sizeof( y_char);
    out.write( reinterpret_cast<char*>( y_int),   sizeof( y_int) );     bytes_counter += sizeof( y_int);
    out.write( reinterpret_cast<char*>( y_float), sizeof( y_float) );   bytes_counter += sizeof( y_float);

    out.write( reinterpret_cast<char*>( &info ), sizeof( info) );       bytes_counter += sizeof( info);

    cout << "we write out " << bytes_counter << " bytes" << endl;

}

void read_binary_file() {

    // Here is also a reminding of how to read binary format.
    fstream in;
	in.open("data.dat" , ios::in|ios::binary);


    char   x_char;
    short  x_short;
    int    x_int;
    float  x_float;
    double x_double;

    char   y_char[2] ={0};
    int    y_int[2]  ={0};
    float  y_float[2]= {0};

    struct info_struct {
        int    z_int;
        float  z_float;
        double z_double[2];
    };

    info_struct info;

    in.read( &x_char, sizeof(x_char) );
    in.read( reinterpret_cast<char*>( &x_short ),  sizeof(x_short) );
    in.read( reinterpret_cast<char*>( &x_int ),    sizeof(x_int) );
    in.read( reinterpret_cast<char*>( &x_float ),  sizeof(x_float) );
    in.read( reinterpret_cast<char*>( &x_double ), sizeof(x_double) );

    in.read( reinterpret_cast<char*>( y_char),  sizeof( y_char) );
    in.read( reinterpret_cast<char*>( y_int),   sizeof( y_int) );
    in.read( reinterpret_cast<char*>( y_float), sizeof( y_float) );

    in.read( reinterpret_cast<char*>( &info ), sizeof( info) );


    cout << " x_char = "  << x_char << endl;
    cout << " x_short = " << x_short << endl;
    cout << " x_int = "   << x_int << endl;
    cout << " x_float = " << x_float << endl;
    cout << " x_double = "<< x_double << endl;

    for( int i=0; i < 2; i++) { cout << "y_char["<<i<<"] = " << y_char[i] << endl; }
    for( int i=0; i < 2; i++) { cout << "y_int["<<i<<"] = "  << y_int[i] << endl;  }
    for( int i=0; i < 2; i++) { cout << "y_float["<<i<<"] = "<< y_float[i] << endl;  }


}

