#include <thread>
#include <iostream>
#include <string>
#include <vector>
#include <mutex>  // for mutual exclusion.
using namespace std;

/*
    if inside takeOut();
    we do ( pop_back() is done some where else..)
    
	int takeOut() {
        std::lock_guard< std::mutex>  guard( m_mutex );
        int x = fVec.back();
        //fVec.pop_back();
        return x;
    }
    this can cause thread-safty issue.
    since pop from thread1 may be come latter than takeOut() by thread2
    so thread2 read the same value as thread1.

*/
 
// a vector<int> with element from 0, 1, 2, ... 99
class myVector{
  private:
 
    std::mutex m_mutex; 
    vector<int> fVec; 
  public:
    myVector() {
        for( int i = 0 ; i<100; i++ ){fVec.push_back( i ); }
    } 

    ~myVector() {  }

    int size() { return fVec.size(); }

    int takeOut() {
        
        std::lock_guard< std::mutex>  guard( m_mutex );

        int x = fVec.back();
        fVec.pop_back();
        return x;
    }



};
 
void take_out_element( myVector& vec, int threadID ){
    
    for( int i=0; i<20; i++) {
        printf("%02d: %02d from t%d thread %2d\n", vec.size(), vec.takeOut(), threadID, i+1 );
    }
    
}

int main() {

    myVector vec;

    // 5 threads, each print 20, so that total = 5 * 20 = 100
    thread* workers[5];
    for( int threadID =0; threadID<5; threadID++ ){
      
        workers[threadID] = new thread( take_out_element, std::ref(vec), threadID+1 );

    }
    
    for( int i =0; i<5; i++ ){
        workers[i]->join();
    }
  
    

    return 0;
}
