#include <thread>
#include <iostream>
#include <string>
#include <mutex>  // for mutual exclusion.
using namespace std;


/* for illustration, let us use cout

    both the main thread and t1 thread chasing to use std::cout.
    thus, the print out will be very messy.

    suppose, we want main thread finish its printing 
    without interference of another printing by t1 thread,
    then we need to lock "printing" for just main thread to use.
    here comes the mutex.
    // you can do mutex::lock() then mutex::unlock()
    // but it is not good, since you may lock it forever by accident,
    a better way it to create an object, and when this object out of scope,
    then mutex automatically unlock it.

    so it comes with td::lock_guard< std::mutex>  guard( mu );
*/



std::mutex mu; // set it as global variable. 

void shared_print( string msg, int iValue ){

    std::lock_guard< std::mutex>  guard( mu );
    // lock it for just one thread to use this function.
    // when guard goes out of scope, then mu is unlock this function
    // for other thread to use.

    cout << msg <<" " << iValue << endl;

    // but it is not perfect,
    // since cout still can be access somewhere else,
    // it is not complete hiden by other threads
}

void Printing_job( ){
    
    for( int i=0; i>-20; i--) {
        shared_print( string("Hello, I am from t1 thread: "), i );    
    }
    
}

int main() {

    thread t1( Printing_job );

     for( int i=0; i<20; i++) {
        shared_print( string("Hey, I am from main thread: "), i );    
    }
    
 
    t1.join();

    return 0;
}
