/*
    demo how to use std::call_once

    we will use LogFile class to write out the info.

    the threads call Printing_job(), in which we use an obj from LogFile class.

    we thus, pass the ref of this obj, "log", into Printing_job() as argument..

    ( in this way, every thread uses just one "log" )

    and so we need to make sure statement like

        f.open( "log2.txt"... ) is just executed once. 

    here comes the help from std::call_once and std::once_flag.

    -- we have to set a global variable for std::once_flag

    -- and then use std::call_once( ) inside the construtor.

********************************************************************/

#include <thread>   // to compile it we need g++ with -pthread
#include <mutex>    // std::call_once, std::once_flag 
#include <iostream>
#include <fstream>

using namespace std;



std::once_flag  _onceFlag;


class LogFile{
  private:
  
    std::mutex m_mutex; 
  
    fstream f;  // we should not let outside to access f
  
  public:
    LogFile() {

        // I use lambda function to do, [&] means I can access
        // this scope, and hence access "f"
        std::call_once( _onceFlag, 
            [&] () {
                f.open("log2.txt", ios::out ); f.close();
                f.open("log2.txt", ios::out );
            } 
        );
    } 

    ~LogFile() { f.close(); }

   

    void shared_print( int threadID, int iValue) {

        // using mutex to lock the process for one thread to finish.
        std::lock_guard< std::mutex>  guard( m_mutex );

        f << "I am from " << threadID << " thead: "<< iValue << endl;
    }
};




// the enter function called by thread.
void Printing_job( LogFile& log, int threadID ){
    
    for( int i=0; i<5; i++) {
        log.shared_print( threadID , i );    
    }
    
}
 


 
int main(){

    LogFile log;

    thread t1( Printing_job, std::ref(log), 1 );
    thread t2( Printing_job, std::ref(log), 2 );
     
    
 
    t1.join();
    t2.join();
    cout << "see log2.txt" << endl;

   
    


    return 0;
}
