/*
    demo how to use std::unique_lock<mutex>
    std::unique_lock< std::mutex > locker( m_mutex, std::defer_lock );
    instead of using lock_guard, we can use unique_lock.
    which can lock and unlock several times, providing more 
    flexibility. However, it reduces the performance. 
    note: std::defer_lock is to tell unique_lock not to lock right away.
    also, 
    unique_lock object can be moved (not copied)
    so we can do:

    std::unique_lock< std::mutex > locker( m_mutex, std::defer_lock );
    std::unique_lock<mutex> locker2 = std::move( locker );

********************************************************************/

#include <thread>   // to compile it we need g++ with -pthread
#include <mutex>    // std::call_once, std::once_flag 
#include <iostream>
#include <fstream>

using namespace std;



std::once_flag  _onceFlag;


class LogFile{
  private:
  
    std::mutex m_mutex; 
  
    fstream f;  // we should not let outside to access f
  
  public:
    LogFile() {

        f.open("log2.txt", ios::out ); f.close();
        f.open("log2.txt", ios::out );           
    } 

    ~LogFile() { f.close(); }

   

    void shared_print( int threadID, int iValue) {

        
        //std::lock_guard< std::mutex>  guard( m_mutex );
        
        std::unique_lock< std::mutex > locker( m_mutex, std::defer_lock );
        // note: std::defer_lock is to tell unique_lock not to lock right away.


        locker.lock();
        f << "I am from " << threadID << " thead: "<< iValue << endl;
        locker.unlock();
        f << "I am from " << threadID << " thead end~~ " << endl;

    }
};




// the enter function called by thread.
void Printing_job( LogFile& log, int threadID ){
    
    for( int i=0; i<5; i++) {
        log.shared_print( threadID , i );    
    }
    
}
 


 
int main(){

    LogFile log;

    thread t1( Printing_job, std::ref(log), 1 );
    thread t2( Printing_job, std::ref(log), 2 );
     
    // call by the main thread.
    for( int i=0; i<5; i++) {
        log.shared_print( 0, 0  );    
    }
 
 
    t1.join();
    t2.join();
    cout << "see log2.txt" << endl;

   
    


    return 0;
}
