#include <iostream>
#include <cmath>
using namespace std;

/*  in this demo, we will have following parts:
    
    part1.1 --- multiple for the memory address
    part1.2 --- array alignment

    part2: test alignas()

    part3: struct

    part4: dynamic

    to compile:
    g++-7 -std=c++1z -o demo 00_demo_alignment.cpp

*/

int maxMultiple( uint64_t input );




void part1_1(){
    // ref: Lecture 8/12: Data Alignment
    // https://www.youtube.com/watch?v=OKjOZBaKlOc
    
    // let's start from the data type with one word, which is 4 bytes (B)
    // reminder 1 byte (B) = 8 bit (b), thus 4 B = 32 b

    uint32_t var_32bit[10];
    for( int i=0; i<10; ++i ){
        cout << &var_32bit[i] << endl;    
    }
    // 0xbfa49fc8
    // 0xbfa49fcc
    // 0xbfa49fd0
    // 0xbfa49fd4
    // 0xbfa49fd8
    // 0xbfa49fdc
    // 0xbfa49fe0
    // 0xbfa49fe4
    // 0xbfa49fe8
    // 0xbfa49fec

    // you can observe the last digit are only 0, 4, 8, c for 32bit's int
    // for hex, c = 12, and it means 

    printf("----------------------------------\n");
    uint64_t var_64bit[10];
    for( int i=0; i<10; ++i ){
        cout << &var_64bit[i] << endl;    
    }
    // 0xbfee6900
    // 0xbfee6908
    // 0xbfee6910
    // 0xbfee6918
    // 0xbfee6920
    // 0xbfee6928
    // 0xbfee6930
    // 0xbfee6938
    // 0xbfee6940
    // 0xbfee6948

    // you can observe the last digit are only 0, 8  for 64bit's int


    /*  a 32bit (=4 bytes=1 word) int's memory address can only 0, 4, 8, 12. why?    
        this is because of aligment rule:
        memory of 4-bytes data type need to be a multiple of 4.
        so it will be easier for CPU to work.
        [    ] 0
        [    ] 1
        [    ] 2
        [    ] 3
        [XXXX] 4  --------
        [XXXX] 5    one
        [XXXX] 6    word
        [XXXX] 7  ---------
        [    ] 8 
        [    ] 9
        [    ] a
        [    ] b
        [XXXX] c
        [XXXX] d
        [XXXX] e
        [XXXX] f
        
        note: each block is a byte.

        for a 64bit (=8 bytes=2 word) int's memory address can only 0 and 8,
        it is from memory of 8-bytes data type need to be a multiple of 8.
        [XXXX] 0
        [XXXX] 1
        [XXXX] 2
        [XXXX] 3
        [XXXX] 4
        [XXXX] 5
        [XXXX] 6
        [XXXX] 7
        [    ] 8 
        [    ] 9
        [    ] a
        [    ] b
        [    ] c
        [    ] d
        [    ] e
        [    ] f

        similarly,
        a 16bit (=2 bytes= 1/2 word) can be at 0, 2, 4, 6, 8, a, c, and e.
        a 8bit  (=1 bytes= 1/4 word) can be anywhere.
    */ 
}

void part1_2(){
    // array alignment
    // https://gain-performance.com/2017/05/07/2629/
    
    // let's start from the data type with one word, which is 4 bytes (B)
    // reminder 1 byte (B) = 8 bit (b), thus 4 B = 32 b

    uint32_t var_32bit[10];
    
    cout <<  "alignof(32bit = 4bytes) = " << alignof( decltype(var_32bit) ) << endl;
    // here we get 4 from alignof() 
    // since var_32bit is 4B, and thus its required alignment is 4B


    cout    << "maxMultiple(32bit = 4bytes) =" 
            << maxMultiple( (uint64_t) &var_32bit ) 
            << endl;
    // the value is not fixed.
    // but it is always larger than its required alignment which is 4B.

    printf("--------------------------------------------\n");

    uint64_t var_64bit[10];
    cout <<  "alignof(64bit = 8bytes) = " <<  alignof( decltype(var_64bit) ) << endl;
    // here we get 8 from alignof() 
    // since var_64bit is 8B, and thus its required alignment is 8B

    cout << "maxMultiple(64bit = 8bytes) =" 
         << maxMultiple( (uint64_t) &var_64bit ) 
         << endl;
    // the value is actuall not fixed.
    // but it is always larger than its required alignment which is 8B.
}

void part1_3(){

    // struct

    struct S1
    {
        uint32_t var_32bit1;
        uint8_t  var_8bit;
        uint32_t var_32bit2;
        // [XXXX] 0 -- uint32_t can only at (0,4,8,c)
        // [XXXX] 1
        // [XXXX] 2
        // [XXXX] 3
        // [XXXX] 4 -- uint8_t (anywhere)
        // [    ] 5
        // [    ] 6
        // [    ] 7
        // [XXXX] 8 -- uint32_t can only at (0,4,8,c) 
        // [XXXX] 9
        // [XXXX] a
        // [XXXX] b
    
    };
    

    struct S2
    {
        uint32_t var_32bit1;
        uint32_t var_32bit2;
        uint8_t  var_8bit;
        // [XXXX] 0 -- uint32_t can only at (0,4,8,c)
        // [XXXX] 1
        // [XXXX] 2
        // [XXXX] 3
        // [XXXX] 4 -- uint32_t can only at (0,4,8,c)
        // [XXXX] 5
        // [XXXX] 6
        // [XXXX] 7
        // [XXXX] 8 -- uint8_t (anywhere)
        // [    ] 9
        // [    ] a
        // [    ] b
    };

    S1 s1;
    S2 s2;

    cout << "[struct] sizeof(s1) = " << sizeof(s1) << " [B]" << endl;
    cout << "[struct] sizeof(s2) = " << sizeof(s2) << " [B]" << endl;
    // the size are the same in my machine and compiler

    cout << "[struct] maxMultiple(s1) = " << maxMultiple( (uint64_t) &s1) << endl;
    cout << "[struct] maxMultiple(s2) = " << maxMultiple( (uint64_t) &s2) << endl;
    // s1 = 8 ( the s1 starting memory only can be 0, 8, 16, 24.... )
    // s2 = 4
}

void part2(){

    // test whether any change if I use alginas()

    alignas(32) uint64_t var_64bit_align32bit[10];

    for( int i=0; i<10; ++i ){
        cout << &var_64bit_align32bit[i] << endl;    
    }
    // 0xbfe92ce0
    // 0xbfe92ce8
    // 0xbfe92cf0
    // 0xbfe92cf8
    // 0xbfe92d00
    // 0xbfe92d08
    // 0xbfe92d10
    // 0xbfe92d18
    // 0xbfe92d20
    // 0xbfe92d28
    // for my machine, there is no difference.

    printf("------------------------------\n");

    alignas(64) uint32_t var_32bit_align64bit[10];

      for( int i=0; i<10; ++i ){
        cout << &var_32bit_align64bit[i] << endl;    
    }
    // 0xbfe92c80
    // 0xbfe92c84
    // 0xbfe92c88
    // 0xbfe92c8c
    // 0xbfe92c90
    // 0xbfe92c94
    // 0xbfe92c98
    // 0xbfe92c9c
    // 0xbfe92ca0
    // 0xbfe92ca4
    // for my machine, there is no difference.
}


int main()
{
    // part1_1();
    
    // part1_2();

    // part1_3();

    part2();
 

    return 0;
}


//=================================================================

/*
    @brief the memory address is a integer number (and it frequently expressed
    in hex format).  we are looking for the max multiple for the address.

    i.e. multiple of 1, 2, 4, 8, 16, 32, .... up to 2048
    so
    memory address = 3,    max multiple = 1 
    memory address = 4,    max multiple = 4 
    memory address = 10,   max multiple = 2  (2*5)
    memory address = 20,   max multiple = 4  (4*5)
    memory address = 32,   max multiple = 32
    memory address = 90,   max multiple = 2  (2*45)
    memory address = 2^20, max multiple = 2048

*/
int maxMultiple( uint64_t input ){

    uint64_t result = 1;
    // note: max = 2048
    for( uint64_t i = 2; i < 4096; i*=2 ) 
    {
        if( input % i != 0) break;
        result = i;
    }

    return result;
}