#include <iostream>
#include <iomanip>
 
void print_address( uint64_t addr ) {

  std::cout << std::showbase // print "0x", meaning hex.
            
            //======================= just fill 0 internally.
            << std::internal 
            << std::setfill('0') 
            //==========================
            <<std::hex
            
            << std::setw(4) 
            
            << addr

            << std::dec;

}

int alignment( uint64_t addr ) {
    // our input is num
    // we search the larget modulo on the base of 2.
    // e.g 18446744072629288079

    uint64_t result = 1;

    for(uint64_t i = 2; i < 4096; i*=2) 
    {
        if(addr % i != 0) break;

        result = i;
    }

    return result;
}
 
using namespace std;

int main()
{
    
    // //            | 8 bytes |  | 8 bytes  |    | 8 bytes |
    // struct Bad  {   char c;      double d;       int i;     }; 
    // cout << alignof(Bad) << " " << sizeof(Bad) << endl;   // 8 24

    // //             | 8 bytes |   |   8 bytes    |    
    // struct Good {   double d;     int i; char c;          };
    // cout << alignof(Good) << " " << sizeof(Good) << endl; // 8 16



    uint8_t a;  // unsigned int with  8 bits.
    uint32_t b; // unsigned int with 16 bits.
    
    // cout <<"-------------------" << endl;
    // printf("alignof a = %d, sizeof a = %d \n", alignof(a), sizeof(a) );


    

    /*
        cout << "32 " << (uint32_t) &a << endl;
        cout << "64 " << (uint64_t) &a << endl;
        32 3221165279
        64 18446744072635749599
          note:
        64bit mean we have something like 101010..... 
        with its lenth = 64. 
        
        it is very long to use binary expression, 
        so people is favor of hex expression,
        1 hex is from 0, 1, 2 ... a, b, c, d, e, f
        total 16 items.
        i.e. 1 hex = 4 bit of binary.

        64bit => having 16 hex 
    */


     
    // note: a is uint8_t, and we convert it to uint64_t
    // we pass by the reference, meaning its mem. addr.
  
    std::cout 
        << "[a]"<< "\naddress: ";  print_address((uint64_t)&a);
    
    std::cout   
        << "\nvariable alignment: "<< alignment((uint64_t)&a)
        << "\nrequired type alignment: " 
        << alignof( decltype(a) ) << "\n";

    // note: decltype( a ) is to deduce the type of varaible a.
    // for alignof see: https://tinyurl.com/y37r47st


  std::cout << "\n[b]"
    << "\naddress: ";
    print_address((uint64_t)&b);
    std::cout << "\nvariable alignment: " << alignment((uint64_t)&b)
    << "\nrequired type alignment: " << alignof(decltype(b)) << "\n";
  
  alignas(16) uint32_t d;
  alignas(4)  uint8_t c;  // be aligned to 4-byte boundary

  std::cout << "\n[c]"
    << "\naddress: ";
    print_address((uint64_t)&c);
    std::cout << "\nvariable alignment: " << alignment((uint64_t)&c)
    << "\nrequired type alignment: " << alignof(decltype(c)) << "\n";
 
  
  std::cout << "\n[d]"
    << "\naddress: ";
    print_address((uint64_t)&d);
    std::cout << "\nvariable alignment: " << alignment((uint64_t)&d)
    << "\nrequired type alignment: " << alignof(decltype(d)) << "\n";



}