/// \file B1ActionInitialization.hh
/// \brief Definition of the B1ActionInitialization class

#ifndef B1ActionInitialization_h
#define B1ActionInitialization_h 1

#include "G4VUserActionInitialization.hh"


/*!
    @brief a top-level class to control all the other user action objects.
    
    + Build(): call SetUserAction() to put in our Run, Event, Stepping objects.
    + BuildForMaster(): setup for for multi-thread 

    the runManager will use this object to control the run. <br>
    runManager->SetUserInitialization(new B1ActionInitialization()); <br>

*/
class B1ActionInitialization : public G4VUserActionInitialization // a must base class
{
  public:

    B1ActionInitialization();

    virtual ~B1ActionInitialization();




    // the virtual methods ( to replace the ones inthe  base class. )

    virtual void BuildForMaster() const; // for the  master thread.

    virtual void Build() const; // for the sequential mode.
    /*

    	here, we will use  SetUserAction() method from G4VUserActionInitialization class

    	to setup the several actions, in this Build() method.

    	1. primary action.
    	2. run action.
    	3. event action
    	4. step action.

        ---------------------------------------------------------
        B1RunAction* runAction = new B1RunAction;
        B1EventAction* eventAction = new B1EventAction(runAction);

        SetUserAction(new B1PrimaryGeneratorAction);
        SetUserAction(runAction);
        SetUserAction(eventAction);
        SetUserAction(new B1SteppingAction(eventAction));

    */



};


#endif


