/// \file B1EventAction.hh
/// \brief Definition of the B1EventAction class

#ifndef B1EventAction_h
#define B1EventAction_h 1

#include "G4UserEventAction.hh"
#include "globals.hh"

class B1RunAction;

/*!
    @brief user event class

    + BeginOfEventAction(): set fEdep = 0 
    + EndOfEventAction(): call AddEdep()
    + AddEdep(): use our "run" obj to update its deposit energy accumulator.
*/
class B1EventAction : public G4UserEventAction
{
  public:
    B1EventAction(B1RunAction* runAction);
    virtual ~B1EventAction();

     
    virtual void BeginOfEventAction(const G4Event* event);
    virtual void EndOfEventAction(const G4Event* event);

    /*!
        @brief for user stepping action object to update the B1EventAction::fEdep.
    */
    void AddEdep(G4double edep) { fEdep += edep; }



  private:

    B1RunAction* fRunAction;

    G4double     fEdep;
};



#endif


