/// @file B1PrimaryGeneratorAction.hh
/// @brief Definition of the B1PrimaryGeneratorAction class

#ifndef B1PrimaryGeneratorAction_h
#define B1PrimaryGeneratorAction_h 1

#include "G4VUserPrimaryGeneratorAction.hh"
#include "G4ParticleGun.hh"
#include "globals.hh"

class G4ParticleGun;
class G4Event;
class G4Box;

/*!
    @brief to set up the beam.
    
    + B1PrimaryGeneratorAction(): set up moving direction, energy for a gamma particle
    + ~B1PrimaryGeneratorAction(): delete fParticleGun 
    + GeneratePrimaries(): setup the position (via random) for our gamma, and generate 
    primary vertex.
    + GetParticleGun(): return our gamma beam.

    in this case: 6MeV gamma, moving toward (0,0,1) direction, <br>
    and 0.8*width and 0.8*height as the area that the gamma will hit
*/
class B1PrimaryGeneratorAction : public G4VUserPrimaryGeneratorAction // a must
{
  public:
    
    B1PrimaryGeneratorAction();            // constructor 
    
    virtual ~B1PrimaryGeneratorAction();   // destructor 


    // method from the base class ( G4VUserPrimaryGeneratorAction class )
    // here we replace it according to our need.
    // ie. 80% of the area.
    virtual void GeneratePrimaries(G4Event*);         
  

    // GetParticleGun method is to access particle gun
    // return type is G4ParticleGun*
    const G4ParticleGun* GetParticleGun() const { return fParticleGun; }
  



  private:
    G4ParticleGun*  fParticleGun; 
    G4Box* fEnvelopeBox;
};


#endif
