/// file B1RunAction.hh
/// brief Definition of the B1RunAction class

#ifndef B1RunAction_h
#define B1RunAction_h 1

#include "G4UserRunAction.hh"
#include "G4Parameter.hh"
#include "globals.hh"

class G4Run;
// it seems be implemented in B1RunAction.cc by #include G4Run.hh


/*!
 @brief Run action class ( run is the biggest unit)

 In BeginOfRunAction(), it sets up the dose unit, <br>
 and Register our fEdep and fEdep2 parameters by parameterManager.<br>

 In EndOfRunAction(), it calculates the dose in the selected volume<br>
 from the energy deposit accumulated via stepping and event actions.<br>
 The computed dose is then printed on the screen.<br>
*/
class B1RunAction : public G4UserRunAction
{
  public:
    B1RunAction();
    virtual ~B1RunAction();

    // virtual G4Run* GenerateRun();

    // BeginOfRunAction() is invoked at the begining of BeamOn() method,
    // and it provides run number identification.
    // EndOfRunAction() is invoked at the end of BeamOn() method.
    virtual void BeginOfRunAction(const G4Run*);

    // It retreive the information ( mass, nofEvents )
    // and calculate the dose, rms, etc..
    // also it print the summary.
    virtual void   EndOfRunAction(const G4Run*);

   
    void AddEdep (G4double edep);

  private:
    G4Parameter<G4double> fEdep;     // Edep = energy deposit
    G4Parameter<G4double> fEdep2;
    // G4Parameter<T> is a templated class,
    // it has the  Reset() method.
    // We will use G4ParamterManager class's RegisterParameter() 
    // to regist the G4Parameter obj.
};

#endif

