/// @file B1ActionInitialization.cc
/// @brief Implementation of the B1ActionInitialization class

#include "B1ActionInitialization.hh"
#include "B1PrimaryGeneratorAction.hh"
#include "B1RunAction.hh"
#include "B1EventAction.hh"
#include "B1SteppingAction.hh"


/// just implement the contructor from the G4VUserActionInitialization class.
B1ActionInitialization::B1ActionInitialization()
 : G4VUserActionInitialization()
{}

/// do nothing
B1ActionInitialization::~B1ActionInitialization()
{}


/// only for the multithread.
/// + create a B1RunAction object "runAction", and put it in SetUserAction()
void B1ActionInitialization::BuildForMaster() const
{
  B1RunAction* runAction = new B1RunAction;
  SetUserAction(runAction);
}



/*!
For the sequential mode.
1. create a B1PrimaryGeneratorAction object, put it in SetUserAction()
2. create a B1RunAction object, put it in SetUserAction()
3. create a B1EventAction object, put it in SetUserAction()
4. create a B1SteppingAction object, put it in SetUserAction()
*/
void B1ActionInitialization::Build() const
{

  SetUserAction(new B1PrimaryGeneratorAction);
  // note: B1PrimaryGeneratorAction class is derived from G4VUserPrimaryGeneratorAction class ( a must )
  // setup: particle gun, with 6 MeV and momentum direction to +z
  // setup: GeneratePrimaries() method, to get 80% of area, associate particle position.


  B1RunAction* runAction = new B1RunAction;
  SetUserAction(runAction);
  //  note: the B1RunAction class derived from G4UserRunAction class
  //  virtual void BeginOfRunAction(const G4Run*); ( reset parameter )
  //  virtual void   EndOfRunAction(const G4Run*); ( print summary )
  //  void AddEdep (G4double edep);



  B1EventAction* eventAction = new B1EventAction(runAction);
  SetUserAction(eventAction);
  //  note: the B1EventAction class derived from G4UserEventAction class
  //  virtual void BeginOfEventAction(const G4Event* event); (to set  fEdep = 0.)
  //  virtual void EndOfEventAction(const G4Event* event);   {to accumulate statistics}
  //  void AddEdep(G4double edep) { fEdep += edep; }



  SetUserAction(new B1SteppingAction(eventAction));
}

