/// \file B1EventAction.cc
/// \brief Implementation of the B1EventAction class

#include "B1EventAction.hh"
#include "B1RunAction.hh"

#include "G4Event.hh"
#include "G4RunManager.hh"


/// B1EventAction::fEdep represents the energy deposit from a given event. <br>
/// here, we initialize B1EventAction::fEdep = 0
B1EventAction::B1EventAction(B1RunAction* runAction)
: G4UserEventAction(),
  fRunAction(runAction),
  fEdep(0.)
{}


/// do nothing.
B1EventAction::~B1EventAction() {}


/// reset B1EventAction::fEdep to 0
void B1EventAction::BeginOfEventAction(const G4Event*) { fEdep = 0.; }


/*!
   at the end of a given event, we use fRunAction object to update
   the B1RunAction::fEdep by call B1RunAction::AddEdep() 
*/
void B1EventAction::EndOfEventAction(const G4Event*)
{
  // accumulate statistics in run action
  // note:
  // B1RunAction* fRunAction;
  // void B1RunAction::AddEdep(G4double edep)
  // { fEdep  += edep; fEdep2 += edep*edep; }
  // the fEdep should be named as fEdep_accumulated
  // the fEdep in Event class, just the energy from a single event.
  fRunAction->AddEdep(fEdep);


}

