/// \file B1SteppingAction.cc
/// \brief Implementation of the B1SteppingAction class

#include "B1SteppingAction.hh"
#include "B1EventAction.hh"
#include "B1DetectorConstruction.hh"

#include "G4Step.hh"
#include "G4Event.hh"
#include "G4RunManager.hh"
#include "G4LogicalVolume.hh"



B1SteppingAction::B1SteppingAction(B1EventAction* eventAction)
    : G4UserSteppingAction(),
      fEventAction(eventAction),
      fScoringVolume(0)
{}



B1SteppingAction::~B1SteppingAction()
{}



// note: UserSteppingAction() is a virtual method.

void B1SteppingAction::UserSteppingAction(const G4Step* step)
{
   // G4Step stores the transient information of a step. 
   // This includes the two endpoints of the step:
   // PreStepPoint and PostStepPoint. G4Step also stores the change in track 
   // properties between the two points. These properties, such as energy and
   // momentum, are updated as the various active processes are invoked.
   // since fScoringVolume is set to 0 initially, !0 = true.
    if (!fScoringVolume)
    {
        // to retrieve our detector setting.
        const B1DetectorConstruction* detectorConstruction
            = static_cast<const B1DetectorConstruction*> 
	    ( G4RunManager::GetRunManager()->GetUserDetectorConstruction() );

        fScoringVolume = detectorConstruction->GetScoringVolume();
        // GetScoringVolume() is get the G4LogicalVolume* logicShape2
        // scoring volume is the one of 80% of the total volume.
    }



    // get volume of the current step;
    // note: step is the input argument.
    G4LogicalVolume* volume
        = step-> GetPreStepPoint()
	      -> GetTouchableHandle()
	      -> GetVolume()
	      -> GetLogicalVolume();

    // check if we are in scoring volume
    if (volume != fScoringVolume) return;

    // collect energy deposited in this step
    G4double edepStep = step->GetTotalEnergyDeposit();
    
    fEventAction->AddEdep(edepStep); 
    // link to the B1EventAction obj.
    // will add the edepStep to the fEep defined in EventAction
}



