import ROOT
import math
import time
import numpy as np
from scipy.integrate import quad
from scipy.optimize import root
from numpy.random import uniform

## we write out the inverse CDF for a normal distribution,
## which can be used for other sampling.

class pdf:
    def __call__( self, t, par ):
        a = par[0] #amplitude
        c = par[1] #center
        s = par[2] #sigma
        x = t[0]   
        tmp = -1.*(x-c)*(x-c)/2. /(s*s)
        return a * math.exp( tmp )


def get_inverse( func, xmin, xmax ):

    grInverseCDF = ROOT.TGraph()

    ipt = 0 
    for xx in np.linspace( xmin, xmax, 1000 ):

        yy = func( xx )
        grInverseCDF.SetPoint( ipt, yy, xx )
        ipt += 1 

    return grInverseCDF, func(xmin), func( xmax )


    

def sampling( func, u ):

    toSolve = lambda x: func(x) - u

    result = root( toSolve, 0. )

    if( result.success ):

        return result.x[0]
    
    else:
        raise ValueError('wired result.')


if __name__ == '__main__':
    

 
    c1 = ROOT.TCanvas( 'c1'  )
    c1.Divide(1, 2, 0)


    #================================ PDF ( prob. dist. function )
    c1.cd(2)
    parN = 3
    func1 = ROOT.TF1("PDF", pdf(), -5, 5, parN )
    func1.SetParameters(1., 0., 1.) # normal dist.
    func1.SetLineColor( ROOT.kBlue )
    func1.GetXaxis().SetRangeUser(-5,5)
    func1.SetLineWidth( 3 )
    func1.Draw( )

    c1.GetPad(2).SetGridx(1)
    c1.GetPad(2).SetGridy(1)


    ##================================ CDF ( cumulative distr. function )
    c1.cd(1)
    func = lambda x:  1. * math.exp(-1.*(x-0.)*(x-0.)/2. /(1.*1.) )
    norm_factor = quad( func, -np.inf,  np.inf )[0]
    pdf_normed = lambda x: func(x)/norm_factor
    
    ## our CDF is here.
    cdf = lambda x: quad( pdf_normed, -np.inf, x )[0]

    gr = ROOT.TGraph(); ipt = 0

    for xx in np.linspace( -5, 5, 100 ):
        gr.SetPoint( ipt, xx, cdf(xx) )
        ipt+=1 
    
    
    gr.Draw("ALP")
    gr.SetLineWidth( 3 )
    gr.SetTitle( "CDF" )
    gr.GetXaxis().SetRangeUser( -5, 5)
    
    c1.GetPad(1).SetGridx(1)
    c1.GetPad(1).SetGridy(1)

    gr.SetLineColor( ROOT.kBlue )
    gr.Draw("ALP")

    
    # c1.cd(2)
    # deriative = lambda x: ( cdf(x+1E-5) - cdf(x) ) / ( 1E-5)
    # gr2 = ROOT.TGraph(); ipt = 0

    # for xx in np.linspace( -5, 5, 100 ):
    #     gr2.SetPoint( ipt, xx, norm_factor*deriative(xx) )
    #     ipt+=1 
    # gr2.Draw("same")
    # gr2.SetLineWidth( 2 )
    # gr2.GetXaxis().SetRangeUser( -5, 5)
    # c1.Close()
    
 

    #==============================================
    # write out inverse CDF function for normal distribution.
    h1 = ROOT.TH1D("h1", "h1", 1000, -5, 5 )
    gr2, ymin, ymax = get_inverse( cdf, -5, 5 )

    outStr = ""
    outStr += "ymin = %15.10e, ymax = %15.10e\n" %( ymin, ymax )

    print "ymin = %15.10e, \nymax = %15.10e\n" %( ymin, ymax )
    for xx in np.linspace( ymin, ymax, 2000 ):
        outStr += "%15.10e %15.10e\n" %( xx, gr2.Eval(xx) )
    
    with open( "coeff.dat", "w" ) as f:
        f.write( outStr )