#include <cmath>
#include <chrono>  // for high_resolution_clock
#include <random>

//========================================= for random generator between [0,1]
static unsigned long jz;
static unsigned long jsr=123456789;
#define SHR3 (jz=jsr, jsr^=(jsr<<13), jsr^=(jsr>>17), jsr^=(jsr<<5),jz+jsr)
#define UNI (.5 + (signed) SHR3*.2328306e-9) 

static unsigned long iz;
//=============================================================================


//========================================= dose engine's random generator
static const double _g_scale1 = 1.0/0x7FFF;
double fastrand01(unsigned int *seed) {
    double a;
    *seed = (214013*(*seed)+2531011);
    a = (double)( ( (*seed) >> 16 ) & 0x7FFF ) * _g_scale1;
    return(a);
}
//=============================================================================


void check_dist_UNI();
void check_dist_fastrand01();
void check_dist_cpp();

//=================================
void test_speed_gRandom();
void test_speed_UNI();
void test_speed_fastrandom01();
void test_speed_cpp();
//=================================

//============================
double one_run_UNI();
double one_run_fastrand01();
void exec_time_test();
//============================ 




void random_generator(){


    test_speed_gRandom();
    test_speed_UNI();
    test_speed_fastrandom01();
    test_speed_cpp();

    
    // check_dist_fastrand01();
    // check_dist_UNI();
    // exec_time_test(); // take long time to run
}





void check_dist_fastrand01(){
    TCanvas* c1 = new TCanvas("c1","fastrand01");
    TH1D* h1_fastrand01 = new TH1D( "h1_fastrand01", "h1_fastrand01", 1000, -0.5, 1.5 );
    unsigned int seed = 0;
    for( int ix = 0; ix<100000; ix++ ){

        double value = fastrand01( &seed ); 
        h1_fastrand01->Fill( value );
    }

    h1_fastrand01->SetLineWidth( 2 );
    h1_fastrand01->SetLineColor( kBlack );
    h1_fastrand01->Draw();
}


void check_dist_UNI(){
    TCanvas* c2 = new TCanvas("c2","UNI");
    TH1D* h1_UNI = new TH1D( "h1_UNI", "h1_UNI", 1000, -0.5, 1.5 );

    for( int ix = 0; ix<100000; ix++ ){

        double value = UNI; 
        h1_UNI->Fill( value );
    }

    h1_UNI->SetLineWidth( 2 );
    h1_UNI->SetLineColor( kBlack );
    h1_UNI->Draw();
}   

void check_dist_cpp(){
    TCanvas* c3 = new TCanvas("c3");
    TH1D* h1 = new TH1D( "h1", "h1", 1000, -0.5, 1.5 );

    std::mt19937 generator (0); // pseudorandom number generator    
    std::uniform_real_distribution<double>  dist(0., 1.);

    for( int ix = 0; ix<100000; ix++ ){

        double value = dist( generator );
        h1->Fill( value );
    }

    h1->SetLineWidth( 2 );
    h1->SetLineColor( kBlack );
    h1->Draw();
}

void exec_time_test(){
   TCanvas* c4 = new TCanvas("c4"); 
   TH1D* h1 = new TH1D( "h1", "h1", 6000, 0, 6000 );
   TH1D* h2 = new TH1D( "h2", "h2", 6000, 0, 6000 );
   for( int ix = 0; ix<20000; ix++ ){

        double value1 = one_run_UNI() * 1E6;; // s -> ms 
        double value2 = one_run_fastrand01() * 1E6;; // s -> ms 

        h1->Fill( value1 );
        h2->Fill( value2 );
   }

   h1->Draw();
   h2->SetLineColor( kRed );
   h2->Draw( "same"   );
}


double one_run_UNI(){

    auto start = std::chrono::high_resolution_clock::now();
    for( int ix = 0; ix < 100000; ix++ ){

        double value = UNI; 
         
    }
    auto finish = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed = finish - start;
    // printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "UNI" );

    return elapsed.count();
}


double one_run_fastrand01(){

    unsigned int seed = 0;
    auto start2 = std::chrono::high_resolution_clock::now();
    for( int ix = 0; ix < 100000; ix++ ){

        double value = fastrand01( &seed ); 
         
    }
    auto finish2 = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed2 = finish2 - start2;
    // printf("Elapsed time: %.3f s (%s)\n", elapsed2.count(), "fastrand01" );

    return elapsed2.count();
}


void test_speed_gRandom(){

    auto start = std::chrono::high_resolution_clock::now();
    for( int ix = 0; ix < 10000000; ix++ ){

        double value = gRandom->Uniform( 0, 1 );
         
    }
    auto finish = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed = finish - start;
    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "gRandom" );
}


void test_speed_UNI(){

    auto start = std::chrono::high_resolution_clock::now();
    for( int ix = 0; ix < 10000000; ix++ ){

        double value = UNI;
         
    }
    auto finish = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed = finish - start;
    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "UNI" );
}


void test_speed_fastrandom01(){

    unsigned int seed = 0;
    auto start = std::chrono::high_resolution_clock::now();
    for( int ix = 0; ix < 10000000; ix++ ){

        double value = fastrand01( &seed ); 
         
    }
    auto finish = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed = finish - start;
    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "fastrand01" );
}



void test_speed_cpp(){

    unsigned int seed = 0;
    auto start = std::chrono::high_resolution_clock::now();


    std::mt19937 generator (0); // pseudorandom number generator    
    std::uniform_real_distribution<double>  dist(0., 1.);

    for( int ix = 0; ix < 10000000; ix++ ){

        double value = dist( generator );
         
    }
    auto finish = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed = finish - start;
    printf("Elapsed time: %.3f s (%s)\n", elapsed.count(), "cpp" );
}