import numpy as np
import ROOT
import time

# this is our probability distribution function.
# just a Gaussian, center at 1, sigma = 2.5
def pdf( x ):
    simga = 2.5;
    cetner = 1;
    PI = ROOT.TMath.Pi()

    term1 = ( x - cetner )**2 / 2. / simga**2 

    # see wiki, and use the PDF form (with 1/sqrt(2 pi sigma))
    return  1./ ROOT.TMath.Sqrt( 2 * PI * simga**2 ) * ROOT.TMath.Exp( -term1 )




# scan the range [x1,x2], and return the max value of func(x)
# scanning step = 1E5
def find_max_pdf( func, x1, x2 ):
    xmin = ROOT.TMath.Min( x1, x2 )
    xmax = ROOT.TMath.Max( x1, x2 )
    xStep = float( xmax - xmin ) / 1E5 
    x = xmin
    ymax = -999
    # search the max value.
    while x <= xmax:
        ytemp = func( x )
        if ytemp > ymax: ymax = ytemp
        x+=xStep
    return ymax


'''
Imagine graphing the density function of a random variable onto a large
rectangular board and throwing darts at it. 

Assume that the darts are uniformly distributed around the board. 
Now remove all of the darts that are outside the area under the curve. 

The remaining darts will be distributed uniformly within the area under the curve,
and the x-positions of these darts will be distributed according to the random variable's density. 
ie. when y=pdf(x_r) is high, it has more opportunity to be larger than y_r.
'''
def sampling( pdf, xmin, xmax, ymin, ymax ):
    
    while( 1 ):
        x_r = ROOT.gRandom.Uniform( xmin, xmax )
    
        y   = pdf( x_r ) 

        y_r = ROOT.gRandom.Uniform( ymin, ymax )

        if( y > y_r ): return x_r
    


if __name__ == '__main__':
    
    xmin = -10.
    xmax =  10.

    # y is the value from pdf(x)
    ymin = 0
    ymax = find_max_pdf( pdf, xmin, xmax )


    # to run and visualize

    h1 = ROOT.TH1F( "h1", "h1", 1000, 1.1*xmin, 1.1*xmax ) # histogram obj.
    
    startTime = time.time()
    for ix in range(1000000):
        xValue = sampling( pdf, xmin, xmax, ymin, ymax  )
        h1.Fill( xValue )
    endTime = time.time()

    print "exec time = ", (endTime - startTime), "seconds"

    c1 = ROOT.TCanvas("c1")
    h1.Draw()
