#include <cmath>
 

//========================================= for random generator between [0,1]
static unsigned long jz;
static unsigned long jsr=123456789;
#define SHR3 (jz=jsr, jsr^=(jsr<<13), jsr^=(jsr>>17), jsr^=(jsr<<5),jz+jsr)
#define UNI (.5 + (signed) SHR3*.2328306e-9) 

//=============================================================================

#define RNOR ( hz=SHR3, iz=hz&127, ( fabs(hz)<kn[iz] )? hz*wn[iz] : nfix() )


static long hz;
static unsigned long iz;

static unsigned long kn[128];
static double wn[128];
static double fn[128]; // used in nfix
 

/*! @brief setting up for normal distribution
    
    @param jsrseed = seed for random generator ( unsigned long )

*/
void zigset( unsigned long jsrseed ){

    
    jsr = jsr ^ jsrseed; // update jsr (the following code do not need jsr)
    // bitwise XOR 
    // (0,0) = (1,1) = 0, while ( 1,0) = (0,1) = 1

    const double m1 = 2147483648.0;
    const double m2 = 4294967296.;
    double dn = 3.442619855899;
    double tn = dn;
    double vn = 9.91256303526217e-3;
    double q;

    
    // Set up tables for RNOR ( normal distribution )
    q = vn / exp( -.5 *dn*dn );
    
    kn[0]= (dn/q) * m1;
    
    kn[1]=0;

    wn[0]=q/m1;

    wn[127]=dn/m1;

    fn[0] =1.;
    
    fn[127] = exp(-.5*dn*dn);

    for(int i=126; i>=1; i-- ) {
        
        dn = sqrt(-2. * log( vn/dn + exp(-.5*dn*dn) )  );
        
        kn[i+1] = (dn/tn) * m1;
        
        tn = dn;

        fn[i] = exp(-.5*dn*dn);
        
        wn[i] = dn/m1;
    }

}


/*! @brief 
    nfix() generates variates from the residue when rejection in RNOR occurs. 
*/
double nfix( void ) {

    const double r = 3.442620;     // The start of the right tail 

    static double x, y;
    
    for( ; ; )    
    {  
        x = hz * wn[iz];        // iz == 0, handles the base strip 
        
        if(iz==0)
        { 
            do{ 
                x = -log(UNI) * 0.2904764;  // .2904764 is 1/r 
                y = -log(UNI);
            }   
            while( y+y<x*x );

            return (hz>0)? r+x : -r-x;
        }
        
        // iz>0, handle the wedges of other strips 
        
        if( fn[iz] + UNI*( fn[iz-1] - fn[iz] ) < exp(-.5*x*x) ) return x;

        // initiate, try to exit for(;;) for loop
        hz=SHR3;
        iz=hz&127;
        if( fabs(hz) <kn[iz] ) return ( hz * wn[iz] );
  }

}

void testFill( );
void printCoeff();

void zig(){

   testFill();
   printCoeff();
}




void testFill( ){

    zigset( 10 ); // setup table i.e. kn, wn, fn.

    double value; 

    TH1D* h1 = new TH1D( "h1", "h1", 1000, -5, 5 );
    
    for( int ix = 0; ix<1000000; ix++ ){

        double value = RNOR;
        h1->Fill( value );
    }

    h1->SetLineWidth( 2 );
    h1->SetLineColor( kBlack );
    h1->Draw();
}

void printCoeff(){

 
    zigset( 10 ); // setup table i.e. kn, wn, fn.

    printf("============kn==========\n");
    int iCount = 0;
    for( int ii = 0; ii< 128; ii++ ) {
        printf( "%ld, ",kn[ii]);  // unsigned long
        iCount += 1;
        if( iCount %4 == 0 ) { printf("\n"); iCount = 0; }

    }
    printf("===========fn===========\n");
    for( int ii = 0; ii< 128; ii++ ) {
        printf( "%.10e, ",fn[ii]);  // double
        iCount += 1;
        if( iCount %4 == 0 ) { printf("\n"); iCount = 0; }
    }
    printf("===========wn===========\n");
    for( int ii = 0; ii< 128; ii++ ) {
        printf( "%.10e, ",wn[ii]);  // double
        iCount += 1;
        if( iCount %4 == 0 ) { printf("\n"); iCount = 0; }
    }

}
 
